# SPDX-FileCopyrightText: Copyright (c) 2024, NVIDIA CORPORATION & AFFILIATES.
# All rights reserved.
# SPDX-License-Identifier: Apache-2.0

# Copyright (c) 2024, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import concurrent.futures
import logging
from typing import Any, Dict, List, Optional, Tuple

import numpy as np
import pandas as pd
import pypdfium2 as libpdfium

from nv_ingest_api.internal.enums.common import ContentTypeEnum
from nv_ingest_api.internal.primitives.nim.default_values import YOLOX_MAX_BATCH_SIZE
from nv_ingest_api.internal.primitives.nim.model_interface.yolox import (
    YOLOX_PAGE_IMAGE_PREPROC_WIDTH,
    YOLOX_PAGE_IMAGE_PREPROC_HEIGHT,
    YoloxPageElementsModelInterface,
    YOLOX_PAGE_IMAGE_FORMAT,
)
from nv_ingest_api.internal.schemas.extract.extract_pdf_schema import PDFiumConfigSchema
from nv_ingest_api.internal.enums.common import TableFormatEnum, TextTypeEnum, AccessLevelEnum
from nv_ingest_api.internal.primitives.nim.model_interface.yolox import (
    YOLOX_PAGE_DEFAULT_VERSION,
    YOLOX_PAGE_CLASS_LABELS,
    get_yolox_page_version,
)
from nv_ingest_api.util.metadata.aggregators import (
    construct_image_metadata_from_base64,
    construct_image_metadata_from_pdf_image,
    extract_pdf_metadata,
    construct_text_metadata,
    construct_page_element_metadata,
    CroppedImageWithContent,
)
from nv_ingest_api.util.nim import create_inference_client
from nv_ingest_api.util.pdf.pdfium import (
    extract_nested_simple_images_from_pdfium_page,
    extract_image_like_objects_from_pdfium_page,
    is_scanned_page,
    pdfium_pages_to_numpy,
)
from nv_ingest_api.util.image_processing import scale_image_to_encoding_size
from nv_ingest_api.util.image_processing.transforms import numpy_to_base64, crop_image

logger = logging.getLogger(__name__)


def _extract_page_elements_using_image_ensemble(
    pages: List[Tuple[int, np.ndarray, Tuple[int, int]]],
    yolox_client,
    execution_trace_log: Optional[List] = None,
) -> List[Tuple[int, object]]:
    """
    Given a list of (page_index, image) tuples and a YOLOX client, this function performs
    inference to extract page element annotations from all pages.

    Parameters
    ----------
    pages : List[Tuple[int, np.ndarray, Tuple[int, int]]]
        List of tuples containing page index, image data as numpy array,
        and optional padding offset information.
    yolox_client : object
        A pre-configured client instance for the YOLOX inference service.
    execution_trace_log : Optional[List], default=None
        List for accumulating execution trace information.

    Returns
    -------
    List[Tuple[int, object]]
        For each page, returns (page_index, joined_content) where joined_content
        is the result of combining annotations from the inference.
    """
    page_elements = []

    try:
        # Collect all page indices and images in order.
        # Optionally, collect padding offsets if present.
        image_page_indices = []
        original_images = []
        padding_offsets = []
        for page in pages:
            image_page_indices.append(page[0])
            original_images.append(page[1])
            if len(pages[0]) > 2:
                padding_offset = page[2]
            else:
                padding_offset = 0
            padding_offsets.append(padding_offset)

        # Prepare the data payload with all images.
        data = {"images": original_images}

        # Perform inference using the NimClient.
        inference_results = yolox_client.infer(
            data,
            model_name="yolox_ensemble",
            max_batch_size=YOLOX_MAX_BATCH_SIZE,
            input_names=["INPUT_IMAGES", "THRESHOLDS"],
            dtypes=["BYTES", "FP32"],
            output_names=["OUTPUT"],
            trace_info=execution_trace_log,
            stage_name="pdf_extraction",
        )

        # Process results: iterate over each image's inference output.
        for annotation_dict, page_index, original_image, padding_offset in zip(
            inference_results, image_page_indices, original_images, padding_offsets
        ):
            _extract_page_element_images(
                annotation_dict,
                original_image,
                page_index,
                page_elements,
                padding_offset,
            )

    except TimeoutError:
        logger.error("Timeout error during page element extraction.")
        raise
    except Exception as e:
        logger.exception(f"Unhandled error during page element extraction: {str(e)}")
        raise

    logger.debug(f"Extracted {len(page_elements)} page elements.")
    return page_elements


# Handle individual page element extraction and model inference
def _extract_page_element_images(
    annotation_dict,
    original_image,
    page_idx,
    page_elements,
    padding_offset=(0, 0),
):
    """
    Handle the extraction of page elements from the inference results and run additional model inference.

    Parameters
    ----------
    annotation_dict : dict/
        A dictionary containing detected objects and their bounding boxes.
    original_image : np.ndarray
        The original image from which objects were detected.
    page_idx : int
        The index of the current page being processed.
    page_elements : List[Tuple[int, ImageTable]]
        A list to which extracted page elements will be appended.

    Notes
    -----
    This function iterates over detected objects, crops the original image to the bounding boxes,
    and runs additional inference on the cropped images to extract detailed information about page
    elements.

    Examples
    --------
    >>> annotation_dict = {"table": [], "chart": []}
    >>> original_image = np.random.rand(1536, 1536, 3)
    >>> page_elements = []
    >>> _extract_page_element_images(annotation_dict, original_image, 0, page_elements)
    """
    orig_width, orig_height, *_ = original_image.shape
    pad_width, pad_height = padding_offset

    if annotation_dict and (set(YOLOX_PAGE_CLASS_LABELS) <= annotation_dict.keys()):
        labels = YOLOX_PAGE_CLASS_LABELS
    else:
        labels = ["table", "chart", "infographics"]

    for label in labels:
        if not annotation_dict:
            continue

        if label not in annotation_dict:
            continue

        objects = annotation_dict[label]

        for idx, bboxes in enumerate(objects):
            *bbox, _ = bboxes
            w1, h1, w2, h2 = bbox

            cropped = crop_image(original_image, (int(w1), int(h1), int(w2), int(h2)))
            if cropped is None:
                continue

            base64_img = numpy_to_base64(cropped, format=YOLOX_PAGE_IMAGE_FORMAT)

            bbox_in_orig_coord = (
                int(w1) - pad_width,
                int(h1) - pad_height,
                int(w2) - pad_width,
                int(h2) - pad_height,
            )
            max_width = orig_width - 2 * pad_width
            max_height = orig_height - 2 * pad_height

            page_element_data = CroppedImageWithContent(
                content="",
                image=base64_img,
                bbox=bbox_in_orig_coord,
                max_width=max_width,
                max_height=max_height,
                type_string=label,
            )
            page_elements.append((page_idx, page_element_data))


def _extract_page_text(page) -> str:
    """
    Always extract text from the given page and return it as a raw string.
    The caller decides whether to use per-page or doc-level logic.
    """
    textpage = page.get_textpage()
    return textpage.get_text_bounded()


def _extract_page_images(
    extract_images_method: str,
    page,
    page_idx: int,
    page_width: float,
    page_height: float,
    page_count: int,
    source_metadata: dict,
    base_unified_metadata: dict,
    **extract_images_params,
) -> list:
    """
    Always extract images from the given page and return a list of image metadata items.
    The caller decides whether to call this based on a flag.
    """
    if extract_images_method == "simple":
        extracted_image_data = extract_nested_simple_images_from_pdfium_page(page)
    else:  # if extract_images_method == "group"
        extracted_image_data = extract_image_like_objects_from_pdfium_page(page, merge=True, **extract_images_params)

    extracted_images = []
    for image_data in extracted_image_data:
        try:
            image_meta = construct_image_metadata_from_pdf_image(
                image_data,
                page_idx,
                page_count,
                source_metadata,
                base_unified_metadata,
            )
            extracted_images.append(image_meta)
        except Exception as e:
            logger.error(f"Unhandled error extracting image on page {page_idx}: {e}")
            # continue extracting other images

    return extracted_images


def _extract_page_elements(
    pages: list,
    page_count: int,
    source_metadata: dict,
    base_unified_metadata: dict,
    extract_tables: bool,
    extract_charts: bool,
    extract_infographics: bool,
    page_to_text_flag_map: Dict[int, bool],
    table_output_format: str,
    yolox_endpoints: Tuple[Optional[str], Optional[str]],
    yolox_infer_protocol: str = "http",
    auth_token: Optional[str] = None,
    execution_trace_log=None,
) -> list:
    """
    Extract page elements from the given pages using YOLOX-based inference.

    This function creates a YOLOX client using the provided parameters, extracts elements
    from pages, and builds metadata for each extracted element based on the specified
    extraction flags.

    Parameters
    ----------
    pages : list
        List of page images to process.
    page_count : int
        Total number of pages in the document.
    source_metadata : dict
        Metadata about the source document.
    base_unified_metadata : dict
        Base metadata to include in all extracted elements.
    extract_tables : bool
        Flag indicating whether to extract tables.
    extract_charts : bool
        Flag indicating whether to extract charts.
    extract_infographics : bool
        Flag indicating whether to extract infographics.
    table_output_format : str
        Format to use for table content.
    yolox_endpoints : Tuple[Optional[str], Optional[str]]
        A tuple containing the gRPC and HTTP endpoints for the YOLOX service.
    yolox_infer_protocol : str, default="http"
        Protocol to use for inference (either "http" or "grpc").
    auth_token : Optional[str], default=None
        Authentication token for the inference service.
    execution_trace_log : optional
        List for accumulating execution trace information.

    Returns
    -------
    list
        List of extracted page elements with their metadata.
    """
    extracted_page_elements = []
    yolox_client = None

    try:
        # Default model name
        yolox_version = YOLOX_PAGE_DEFAULT_VERSION

        # Get the HTTP endpoint to determine the model name if needed
        yolox_http_endpoint = yolox_endpoints[1]
        if yolox_http_endpoint:
            try:
                yolox_version = get_yolox_page_version(yolox_http_endpoint)
            except Exception as e:
                logger.warning(f"Failed to get YOLOX model name from endpoint: {e}. Using default.")

        # Create the model interface
        model_interface = YoloxPageElementsModelInterface(version=yolox_version)
        # Create the inference client
        yolox_client = create_inference_client(
            yolox_endpoints,
            model_interface,
            auth_token,
            yolox_infer_protocol,
        )

        # Extract page elements using the client
        page_element_results = _extract_page_elements_using_image_ensemble(
            pages, yolox_client, execution_trace_log=execution_trace_log
        )

        # Process each extracted element based on extraction flags
        for page_idx, page_element in page_element_results:
            process_text_for_this_page = page_to_text_flag_map.get(page_idx, False)
            element_type = page_element.type_string

            page_reading_index = page_idx + 1

            # Skip elements that shouldn't be extracted based on flags
            if (not extract_tables) and (element_type == "table"):
                continue
            if (not extract_charts) and (element_type == "chart"):
                continue
            if (not extract_infographics) and (element_type == "infographic"):
                continue
            if (not process_text_for_this_page) and (element_type in {"title", "paragraph", "header_footer"}):
                continue

            # Set content format for tables
            if page_element.type_string == "table":
                page_element.content_format = table_output_format

            # Construct metadata for the page element
            page_element_meta = construct_page_element_metadata(
                page_element,
                page_reading_index,
                page_count,
                source_metadata,
                base_unified_metadata,
            )
            extracted_page_elements.append(page_element_meta)

    except Exception as e:
        logger.exception(f"Error in page element extraction: {str(e)}")
        raise

    return extracted_page_elements


def pdfium_extractor(
    pdf_stream,
    extract_text: bool,
    extract_images: bool,
    extract_infographics: bool,
    extract_tables: bool,
    extract_charts: bool,
    extract_page_as_image: bool,
    extractor_config: dict,
    execution_trace_log: Optional[List[Any]] = None,
) -> pd.DataFrame:
    # --- Extract and validate extractor_config ---
    if extractor_config is None or not isinstance(extractor_config, dict):
        raise ValueError("`extractor_config` must be provided as a dictionary.")

    # Validate and extract row_data
    row_data = extractor_config.get("row_data")
    if row_data is None:
        raise ValueError("`extractor_config` must include a valid 'row_data' dictionary.")
    if "source_id" not in row_data:
        raise ValueError("The 'row_data' dictionary must contain the 'source_id' key.")

    # Validate and extract text_depth
    text_depth_str = extractor_config.get("text_depth", "page")
    try:
        text_depth = TextTypeEnum[text_depth_str.upper()]
    except KeyError:
        raise ValueError(
            f"Invalid text_depth: {text_depth_str}. Valid options: {list(TextTypeEnum.__members__.keys())}"
        )

    # Validate and extract table_output_format
    table_output_format_str = extractor_config.get("table_output_format", "pseudo_markdown")
    try:
        table_output_format = TableFormatEnum[table_output_format_str.upper()]
    except KeyError:
        raise ValueError(
            f"Invalid table_output_format: {table_output_format_str}. "
            f"Valid options: {list(TableFormatEnum.__members__.keys())}"
        )

    text_extraction_method = extractor_config.get("extract_method", "pdfium")
    extract_images_method = extractor_config.get("extract_images_method", "group")
    extract_images_params = extractor_config.get("extract_images_params", {})

    # Extract metadata_column
    metadata_column = extractor_config.get("metadata_column", "metadata")

    # Process pdfium_config
    pdfium_config_raw = extractor_config.get("pdfium_config", {})
    if isinstance(pdfium_config_raw, dict):
        pdfium_config = PDFiumConfigSchema(**pdfium_config_raw)
    elif isinstance(pdfium_config_raw, PDFiumConfigSchema):
        pdfium_config = pdfium_config_raw
    else:
        raise ValueError("`pdfium_config` must be a dictionary or a PDFiumConfigSchema instance.")
    # --- End extractor_config extraction ---

    logger.debug("Extracting PDF with pdfium backend.")
    source_id = row_data["source_id"]

    # Retrieve unified metadata robustly (supporting pandas Series or dict)
    if hasattr(row_data, "index"):
        base_unified_metadata = row_data[metadata_column] if metadata_column in row_data.index else {}
    else:
        base_unified_metadata = row_data.get(metadata_column, {})

    base_source_metadata = base_unified_metadata.get("source_metadata", {})
    source_location = base_source_metadata.get("source_location", "")
    collection_id = base_source_metadata.get("collection_id", "")
    partition_id = base_source_metadata.get("partition_id", -1)
    access_level = base_source_metadata.get("access_level", AccessLevelEnum.UNKNOWN)

    doc = libpdfium.PdfDocument(pdf_stream)
    pdf_metadata = extract_pdf_metadata(doc, source_id)
    page_count = pdf_metadata.page_count

    source_metadata = {
        "source_name": pdf_metadata.filename,
        "source_id": source_id,
        "source_location": source_location,
        "source_type": pdf_metadata.source_type,
        "collection_id": collection_id,
        "date_created": pdf_metadata.date_created,
        "last_modified": pdf_metadata.last_modified,
        "summary": "",
        "partition_id": partition_id,
        "access_level": access_level,
    }

    logger.debug(f"PDF has {page_count} pages.")
    logger.debug(
        f"extract_text={extract_text}, extract_images={extract_images}, "
        f"extract_tables={extract_tables}, extract_charts={extract_charts}, "
        f"extract_infographics={extract_infographics}"
    )

    # Decide if text extraction should be done at the PAGE or DOCUMENT level
    if text_depth != TextTypeEnum.PAGE:
        text_depth = TextTypeEnum.DOCUMENT

    extracted_data = []
    accumulated_text = []

    # Prepare for table/chart/infographic/text OCR extraction
    pages_for_extractions = []  # Accumulate tuples of (page_idx, np_image)
    page_to_text_flag_map = {}  # Maps page_idx -> bool (True if OCR text extraction is needed)
    futures = []  # To track asynchronous table/chart extraction tasks

    with concurrent.futures.ThreadPoolExecutor(max_workers=pdfium_config.workers_per_progress_engine) as executor:
        # PAGE LOOP
        for page_idx in range(page_count):
            page = doc.get_page(page_idx)
            page_width, page_height = page.get_size()
            page_reading_index = page_idx + 1

            is_scanned = is_scanned_page(page)
            extraction_needed_for_text = extract_text and (
                (text_extraction_method == "pdfium_hybrid" and is_scanned) or text_extraction_method == "ocr"
            )
            extraction_needed_for_structured = extract_tables or extract_charts or extract_infographics
            page_to_text_flag_map[page_idx] = extraction_needed_for_text

            # Text extraction
            if extract_text and not extraction_needed_for_text:
                page_text = _extract_page_text(page)
                if text_depth == TextTypeEnum.PAGE:
                    text_meta = construct_text_metadata(
                        [page_text],
                        pdf_metadata.keywords,
                        page_reading_index,
                        -1,
                        -1,
                        -1,
                        page_count,
                        text_depth,
                        source_metadata,
                        base_unified_metadata,
                    )
                    extracted_data.append(text_meta)
                else:
                    accumulated_text.append(page_text)

            # Image extraction
            if extract_images:
                image_data = _extract_page_images(
                    extract_images_method,
                    page,
                    page_reading_index,
                    page_width,
                    page_height,
                    page_count,
                    source_metadata,
                    base_unified_metadata,
                    **extract_images_params,
                )
                extracted_data.extend(image_data)

            # Full page image extraction
            if extract_page_as_image:
                if text_extraction_method == "ocr":
                    page_text = ""
                else:
                    page_text = _extract_page_text(page)
                image, _ = pdfium_pages_to_numpy([page], scale_tuple=(16384, 16384), trace_info=execution_trace_log)
                base64_image = numpy_to_base64(image[0])
                if len(base64_image) > 2**24 - 1:
                    base64_image, _ = scale_image_to_encoding_size(base64_image, max_base64_size=2**24 - 1)
                image_meta = construct_image_metadata_from_base64(
                    base64_image,
                    page_reading_index,
                    page_count,
                    source_metadata,
                    base_unified_metadata,
                    subtype=ContentTypeEnum.PAGE_IMAGE,
                    text=page_text,
                )
                extracted_data.append(image_meta)

            # If we want OCR extraction, rasterize the page and store it
            if extraction_needed_for_text or extraction_needed_for_structured:
                image, padding_offsets = pdfium_pages_to_numpy(
                    [page],
                    scale_tuple=(YOLOX_PAGE_IMAGE_PREPROC_WIDTH, YOLOX_PAGE_IMAGE_PREPROC_HEIGHT),
                    padding_tuple=(YOLOX_PAGE_IMAGE_PREPROC_WIDTH, YOLOX_PAGE_IMAGE_PREPROC_HEIGHT),
                    trace_info=execution_trace_log,
                )
                pages_for_extractions.append((page_idx, image[0], padding_offsets[0]))

                # Whenever pages_for_extractions hits YOLOX_MAX_BATCH_SIZE, submit a job
                if len(pages_for_extractions) >= YOLOX_MAX_BATCH_SIZE:
                    future = executor.submit(
                        _extract_page_elements,
                        pages_for_extractions[:],  # pass a copy
                        page_count,
                        source_metadata,
                        base_unified_metadata,
                        extract_tables=extract_tables,
                        extract_charts=extract_charts,
                        extract_infographics=extract_infographics,
                        page_to_text_flag_map=page_to_text_flag_map,
                        table_output_format=table_output_format,
                        yolox_endpoints=pdfium_config.yolox_endpoints,
                        yolox_infer_protocol=pdfium_config.yolox_infer_protocol,
                        auth_token=pdfium_config.auth_token,
                        execution_trace_log=execution_trace_log,
                    )
                    futures.append(future)
                    pages_for_extractions.clear()

            page.close()

        # After page loop, if we still have leftover pages_for_extractions, submit one last job
        if (extraction_needed_for_text or extraction_needed_for_structured) and pages_for_extractions:
            future = executor.submit(
                _extract_page_elements,
                pages_for_extractions[:],
                page_count,
                source_metadata,
                base_unified_metadata,
                extract_tables=extract_tables,
                extract_charts=extract_charts,
                extract_infographics=extract_infographics,
                page_to_text_flag_map=page_to_text_flag_map,
                table_output_format=table_output_format,
                yolox_endpoints=pdfium_config.yolox_endpoints,
                yolox_infer_protocol=pdfium_config.yolox_infer_protocol,
                auth_token=pdfium_config.auth_token,
                execution_trace_log=execution_trace_log,
            )
            futures.append(future)

            pages_for_extractions.clear()

        # Wait for all asynchronous jobs to complete.
        for fut in concurrent.futures.as_completed(futures):
            table_chart_items = fut.result()  # Blocks until the job is finished
            extracted_data.extend(table_chart_items)

    # For document-level text extraction, combine the accumulated text.
    if extract_text and text_depth == TextTypeEnum.DOCUMENT and accumulated_text:
        doc_text_meta = construct_text_metadata(
            accumulated_text,
            pdf_metadata.keywords,
            -1,
            -1,
            -1,
            -1,
            page_count,
            text_depth,
            source_metadata,
            base_unified_metadata,
        )
        extracted_data.append(doc_text_meta)

    doc.close()

    logger.debug(f"Extracted {len(extracted_data)} items from PDF.")
    return extracted_data



# SPDX-FileCopyrightText: Copyright (c) 2024, NVIDIA CORPORATION & AFFILIATES.
# All rights reserved.
# SPDX-License-Identifier: Apache-2.0


from pydantic import ConfigDict, BaseModel, model_validator, field_validator, Field


class ImageCaptionExtractionSchema(BaseModel):
    api_key: str = Field(default="", repr=False)
    endpoint_url: str = "https://integrate.api.nvidia.com/v1/chat/completions"
    prompt: str = "Caption the content of this image:"
    model_name: str = "nvidia/nemotron-nano-12b-v2-vl"
    raise_on_failure: bool = False
    model_config = ConfigDict(extra="forbid")

    @field_validator("api_key", mode="before")
    @classmethod
    def _coerce_api_key_none(cls, v):
        return "" if v is None else v

    @model_validator(mode="before")
    @classmethod
    def _coerce_none_to_empty(cls, values):
        """Allow None for string fields where empty string is acceptable.

        Specifically, convert api_key=None to api_key="" so validation passes
        when no API key is supplied.
        """
        if isinstance(values, dict):
            if values.get("api_key") is None:
                values["api_key"] = ""
        return values



# SPDX-FileCopyrightText: Copyright (c) 2024-25, NVIDIA CORPORATION & AFFILIATES.
# All rights reserved.
# SPDX-License-Identifier: Apache-2.0

import logging
from concurrent.futures import ThreadPoolExecutor
from functools import partial
from typing import Any, Dict, Tuple, Optional, Iterable, List
from urllib.parse import urlparse

import glom
import pandas as pd

from nv_ingest_api.internal.enums.common import ContentTypeEnum
from nv_ingest_api.internal.schemas.transform.transform_text_embedding_schema import TextEmbeddingSchema
from nv_ingest_api.util.nim import infer_microservice


logger = logging.getLogger(__name__)

# Reduce SDK HTTP logging verbosity so request/response logs are not emitted
logging.getLogger("httpx").setLevel(logging.ERROR)
logging.getLogger("httpcore").setLevel(logging.ERROR)


MULTI_MODAL_MODELS = ["llama-3.2-nemoretriever-1b-vlm-embed-v1"]


# ------------------------------------------------------------------------------
# Asynchronous Embedding Requests
# ------------------------------------------------------------------------------


def _make_async_request(
    prompts: List[str],
    api_key: str,
    embedding_nim_endpoint: str,
    embedding_model: str,
    encoding_format: str,
    input_type: str,
    truncate: str,
    filter_errors: bool,
    modalities: Optional[List[str]] = None,
    dimensions: Optional[int] = None,
) -> list:
    """
    Interacts directly with the NIM embedding service to calculate embeddings for a batch of prompts.

    Parameters
    ----------
    prompts : List[str]
        A list of prompt strings for which embeddings are to be calculated.
    api_key : str
        API key for authentication with the embedding service.
    embedding_nim_endpoint : str
        Base URL for the NIM embedding service.
    embedding_model : str
        The model to use for generating embeddings.
    encoding_format : str
        The desired encoding format.
    input_type : str
        The type of input data.
    truncate : str
        Truncation setting for the input data.
    filter_errors : bool
        Flag indicating whether to filter errors in the response.

    Returns
    -------
    list
        A dictionary with keys "embedding" (the embedding results) and "info_msg" (any error info).

    Raises
    ------
    RuntimeError
        If an error occurs during the embedding request, with an info message attached.
    """
    response = {}

    try:
        # Normalize API key to avoid sending an empty bearer token via SDK internals
        _token = (api_key or "").strip()
        _api_key = _token if _token else "<no key provided>"

        resp = infer_microservice(
            prompts,
            embedding_model,
            embedding_endpoint=embedding_nim_endpoint,
            nvidia_api_key=_api_key,
            input_type=input_type,
            truncate=truncate,
            batch_size=8191,
            grpc="http" not in urlparse(embedding_nim_endpoint).scheme,
            input_names=["text"],
            output_names=["embeddings"],
            dtypes=["BYTES"],
        )

        response["embedding"] = resp
        response["info_msg"] = None

    except Exception as err:
        # Truncate error message to prevent memory blowup from large text content
        err_str = str(err)
        if len(err_str) > 500:
            truncated_err = err_str[:200] + "... [truncated to prevent memory blowup] ..." + err_str[-100:]
        else:
            truncated_err = err_str

        raise RuntimeError(f"Embedding error occurred: {truncated_err}") from err

    return response


def _async_request_handler(
    prompts: List[str],
    api_key: str,
    embedding_nim_endpoint: str,
    embedding_model: str,
    encoding_format: str,
    input_type: str,
    truncate: str,
    filter_errors: bool,
    modalities: Optional[List[str]] = None,
    dimensions: Optional[int] = None,
) -> List[dict]:
    """
    Gathers calculated embedding results from the NIM embedding service concurrently.

    Parameters
    ----------
    prompts : List[str]
        A list of prompt batches.
    api_key : str
        API key for authentication.
    embedding_nim_endpoint : str
        Base URL for the NIM embedding service.
    embedding_model : str
        The model to use for generating embeddings.
    encoding_format : str
        The desired encoding format.
    input_type : str
        The type of input data.
    truncate : str
        Truncation setting for the input data.
    filter_errors : bool
        Flag indicating whether to filter errors in the response.

    Returns
    -------
    List[dict]
        A list of response dictionaries from the embedding service.
    """
    if modalities is None:
        modalities = [None] * len(prompts)

    with ThreadPoolExecutor() as executor:
        futures = [
            executor.submit(
                _make_async_request,
                prompts=prompt_batch,
                api_key=api_key,
                embedding_nim_endpoint=embedding_nim_endpoint,
                embedding_model=embedding_model,
                encoding_format=encoding_format,
                input_type=input_type,
                truncate=truncate,
                filter_errors=filter_errors,
                modalities=modality_batch,
                dimensions=dimensions,
            )
            for prompt_batch, modality_batch in zip(prompts, modalities)
        ]
        results = [future.result() for future in futures]

    return results


def _async_runner(
    prompts: List[str],
    api_key: str,
    embedding_nim_endpoint: str,
    embedding_model: str,
    encoding_format: str,
    input_type: str,
    truncate: str,
    filter_errors: bool,
    modalities: Optional[List[str]] = None,
    dimensions: Optional[int] = None,
) -> dict:
    """
    Concurrently launches all NIM embedding requests and flattens the results.

    Parameters
    ----------
    prompts : List[str]
        A list of prompt batches.
    api_key : str
        API key for authentication.
    embedding_nim_endpoint : str
        Base URL for the NIM embedding service.
    embedding_model : str
        The model to use for generating embeddings.
    encoding_format : str
        The desired encoding format.
    input_type : str
        The type of input data.
    truncate : str
        Truncation setting for the input data.
    filter_errors : bool
        Flag indicating whether to filter errors in the response.

    Returns
    -------
    dict
        A dictionary with keys "embeddings" (flattened embedding results) and "info_msgs" (error messages).
    """
    results = _async_request_handler(
        prompts,
        api_key,
        embedding_nim_endpoint,
        embedding_model,
        encoding_format,
        input_type,
        truncate,
        filter_errors,
        modalities=modalities,
        dimensions=dimensions,
    )

    flat_results = {"embeddings": [], "info_msgs": []}
    for batch_dict in results:
        info_msg = batch_dict["info_msg"]
        for embedding in batch_dict["embedding"]:
            if not isinstance(embedding, list):
                if embedding is not None:
                    flat_results["embeddings"].append(embedding.embedding)
                else:
                    flat_results["embeddings"].append(embedding)
            else:
                flat_results["embeddings"].append(embedding)
            flat_results["info_msgs"].append(info_msg)

    return flat_results


# ------------------------------------------------------------------------------
# Pandas UDFs for Content Extraction
# ------------------------------------------------------------------------------


def _add_embeddings(row, embeddings, info_msgs):
    """
    Updates a DataFrame row with embedding data and associated error info.
    Ensures the 'embedding' field is always present, even if None.

    Parameters
    ----------
    row : pandas.Series
        A row of the DataFrame.
    embeddings : dict
        Dictionary mapping row indices to embeddings.
    info_msgs : dict
        Dictionary mapping row indices to info message dicts.

    Returns
    -------
    pandas.Series
        The updated row with 'embedding', 'info_message_metadata', and
        '_contains_embeddings' appropriately set.
    """
    embedding = embeddings.get(row.name, None)
    info_msg = info_msgs.get(row.name, None)

    # Always set embedding, even if None
    row["metadata"]["embedding"] = embedding

    if info_msg:
        row["metadata"]["info_message_metadata"] = info_msg
        row["document_type"] = ContentTypeEnum.INFO_MSG
        row["_contains_embeddings"] = False
    else:
        row["_contains_embeddings"] = embedding is not None

    return row


def _add_custom_embeddings(row, embeddings, result_target_field):
    """
    Updates a DataFrame row with embedding data and associated error info
    based on a user supplied custom content field.

    Parameters
    ----------
    row : pandas.Series
        A row of the DataFrame.
    embeddings : dict
        Dictionary mapping row indices to embeddings.
    result_target_field: str
        The field in custom_content to output the embeddings to

    Returns
    -------
    pandas.Series
        The updated row
    """
    embedding = embeddings.get(row.name, None)

    if embedding is not None:
        row["metadata"] = glom.assign(row["metadata"], "custom_content." + result_target_field, embedding, missing=dict)

    return row


def _format_image_input_string(image_b64: Optional[str]) -> str:
    if not image_b64:
        return
    return f"data:image/png;base64,{image_b64}"


def _format_text_image_pair_input_string(text: Optional[str], image_b64: Optional[str]) -> str:
    if (not text) or (not text.strip()) or (not image_b64):
        return
    return f"{text.strip()} {_format_image_input_string(image_b64)}"


def _get_pandas_text_content(row, modality="text"):
    """
    Extracts text content from a DataFrame row.

    Parameters
    ----------
    row : pandas.Series
        A row containing the 'content' key.

    Returns
    -------
    str
        The text content from the row.
    """
    return row["content"]


def _get_pandas_table_content(row, modality="text"):
    """
    Extracts table/chart content from a DataFrame row.

    Parameters
    ----------
    row : pandas.Series
        A row containing 'table_metadata' with 'table_content'.

    Returns
    -------
    str
        The table/chart content from the row.
    """
    if modality == "text":
        content = row.get("table_metadata", {}).get("table_content")
    elif modality == "image":
        content = _format_image_input_string(row.get("content"))
    elif modality == "text_image":
        text = row.get("table_metadata", {}).get("table_content")
        image = row.get("content")
        content = _format_text_image_pair_input_string(text, image)

    return content


def _get_pandas_image_content(row, modality="text"):
    """
    Extracts image caption content from a DataFrame row.

    Parameters
    ----------
    row : pandas.Series
        A row containing 'image_metadata' with 'caption'.

    Returns
    -------
    str
        The image caption from the row.
    """
    subtype = row.get("content_metadata", {}).get("subtype")
    if modality == "text":
        if subtype == "page_image":
            content = row.get("image_metadata", {}).get("text")
        else:
            content = row.get("image_metadata", {}).get("caption")
    elif modality == "image":
        content = _format_image_input_string(row.get("content"))
    elif modality == "text_image":
        if subtype == "page_image":
            text = row.get("image_metadata", {}).get("text")
        else:
            text = row.get("image_metadata", {}).get("caption")
        image = row.get("content")
        content = _format_text_image_pair_input_string(text, image)

    if subtype == "page_image":
        # A workaround to save memory for full page images.
        row["content"] = ""

    return content


def _get_pandas_audio_content(row, modality="text"):
    """
    A pandas UDF used to select extracted audio transcription to be used to create embeddings.
    """
    return row.get("audio_metadata", {}).get("audio_transcript")


def _get_pandas_custom_content(row, custom_content_field):
    custom_content = row.get("custom_content", {})
    content = glom.glom(custom_content, custom_content_field, default=None)
    if content is None:
        logger.warning(f"Custom content field: {custom_content_field} not found")
        return None

    try:
        return str(content)
    except (TypeError, ValueError):
        logger.warning(f"Cannot convert custom content field: {custom_content_field} to string")
        return None


# ------------------------------------------------------------------------------
# Batch Processing Utilities
# ------------------------------------------------------------------------------


def _batch_generator(iterable: Iterable, batch_size: int = 10):
    """
    Yields batches of a specified size from an iterable.

    Parameters
    ----------
    iterable : Iterable
        The iterable to batch.
    batch_size : int, optional
        The size of each batch (default is 10).

    Yields
    ------
    list
        A batch of items from the iterable.
    """
    iter_len = len(iterable)
    for idx in range(0, iter_len, batch_size):
        yield iterable[idx : min(idx + batch_size, iter_len)]


def _generate_batches(prompts: List[str], batch_size: int = 100) -> List[str]:
    """
    Splits a list of prompts into batches.

    Parameters
    ----------
    prompts : List[str]
        The list of prompt strings.
    batch_size : int, optional
        The desired batch size (default is 100).

    Returns
    -------
    List[List[str]]
        A list of batches, each containing a subset of the prompts.
    """
    return [batch for batch in _batch_generator(prompts, batch_size)]


# ------------------------------------------------------------------------------
# DataFrame Concatenation Utility
# ------------------------------------------------------------------------------


def _concatenate_extractions_pandas(
    base_df: pd.DataFrame, dataframes: List[pd.DataFrame], masks: List[pd.Series]
) -> pd.DataFrame:
    """
    Concatenates processed DataFrame rows (with embeddings) with unprocessed rows from the base DataFrame.

    Parameters
    ----------
    base_df : pd.DataFrame
        The original DataFrame.
    dataframes : List[pd.DataFrame]
        List of DataFrames that have been enriched with embeddings.
    masks : List[pd.Series]
        List of boolean masks indicating the rows that were processed.

    Returns
    -------
    pd.DataFrame
        The concatenated DataFrame with embeddings applied where available.
    """
    unified_mask = pd.Series(False, index=base_df.index)
    for mask in masks:
        unified_mask = unified_mask | mask

    df_no_text = base_df.loc[~unified_mask].copy()
    df_no_text["_contains_embeddings"] = False

    dataframes.append(df_no_text)
    combined_df = pd.concat(dataframes, axis=0, ignore_index=True).reset_index(drop=True)
    return combined_df


# ------------------------------------------------------------------------------
# Embedding Extraction Pipeline
# ------------------------------------------------------------------------------


def does_model_support_multimodal_embeddings(model: str) -> bool:
    """
    Checks if a given model supports multi-modal embeddings.

    Parameters
    ----------
    model : str
        The name of the model.

    Returns
    -------
    bool
        True if the model supports multi-modal embeddings, False otherwise.
    """
    return model in MULTI_MODAL_MODELS


def transform_create_text_embeddings_internal(
    df_transform_ledger: pd.DataFrame,
    task_config: Dict[str, Any],
    transform_config: TextEmbeddingSchema = TextEmbeddingSchema(),
    execution_trace_log: Optional[Dict] = None,
) -> Tuple[pd.DataFrame, Dict]:
    """
    Generates text embeddings for supported content types (TEXT, STRUCTURED, IMAGE, AUDIO)
    from a pandas DataFrame using asynchronous requests.

    This function ensures that even if the extracted content is empty or None,
    the embedding field is explicitly created and set to None.

    Parameters
    ----------
    df_transform_ledger : pd.DataFrame
        The DataFrame containing content for embedding extraction.
    task_config : Dict[str, Any]
        Dictionary containing task properties (e.g., filter error flag).
    transform_config : TextEmbeddingSchema, optional
        Validated configuration for text embedding extraction.
    execution_trace_log : Optional[Dict], optional
        Optional trace information for debugging or logging (default is None).

    Returns
    -------
    Tuple[pd.DataFrame, Dict]
        A tuple containing:
            - The updated DataFrame with embeddings applied.
            - A dictionary with trace information.
    """
    api_key = task_config.get("api_key") or transform_config.api_key
    endpoint_url = task_config.get("endpoint_url") or transform_config.embedding_nim_endpoint
    model_name = task_config.get("model_name") or transform_config.embedding_model
    custom_content_field = task_config.get("custom_content_field") or transform_config.custom_content_field
    dimensions = task_config.get("dimensions") or transform_config.dimensions

    if execution_trace_log is None:
        execution_trace_log = {}
        logger.debug("No trace_info provided. Initialized empty trace_info dictionary.")

    if df_transform_ledger.empty:
        return df_transform_ledger, {"trace_info": execution_trace_log}

    embedding_dataframes = []
    content_masks = []

    pandas_content_extractor = {
        ContentTypeEnum.TEXT: _get_pandas_text_content,
        ContentTypeEnum.STRUCTURED: _get_pandas_table_content,
        ContentTypeEnum.IMAGE: _get_pandas_image_content,
        ContentTypeEnum.AUDIO: _get_pandas_audio_content,
        ContentTypeEnum.VIDEO: lambda x: None,  # Not supported yet.
    }
    task_type_to_modality = {
        ContentTypeEnum.TEXT: task_config.get("text_elements_modality") or transform_config.text_elements_modality,
        ContentTypeEnum.STRUCTURED: (
            task_config.get("structured_elements_modality") or transform_config.structured_elements_modality
        ),
        ContentTypeEnum.IMAGE: task_config.get("image_elements_modality") or transform_config.image_elements_modality,
        ContentTypeEnum.AUDIO: task_config.get("audio_elements_modality") or transform_config.audio_elements_modality,
        ContentTypeEnum.VIDEO: lambda x: None,  # Not supported yet.
    }

    def _content_type_getter(row):
        return row["content_metadata"]["type"]

    for content_type, content_getter in pandas_content_extractor.items():
        if not content_getter:
            logger.warning(f"Skipping text_embedding generation for unsupported content type: {content_type}")
            continue

        # Get rows matching the content type
        content_mask = df_transform_ledger["metadata"].apply(_content_type_getter) == content_type.value
        if not content_mask.any():
            continue

        # Always include all content_mask rows and prepare them
        df_content = df_transform_ledger.loc[content_mask].copy().reset_index(drop=True)

        # Extract content and normalize empty or non-str to None
        extracted_content = (
            df_content["metadata"]
            .apply(partial(content_getter, modality=task_type_to_modality[content_type]))
            .apply(lambda x: x.strip() if isinstance(x, str) and x.strip() else None)
        )
        df_content["_content"] = extracted_content

        # Prepare batches for only valid (non-None) content
        valid_content_mask = df_content["_content"].notna()
        if valid_content_mask.any():
            filtered_content_list = df_content.loc[valid_content_mask, "_content"].tolist()
            filtered_content_batches = _generate_batches(filtered_content_list, batch_size=transform_config.batch_size)

            if model_name in MULTI_MODAL_MODELS:
                modality_list = [task_type_to_modality[content_type]] * len(filtered_content_list)
                modality_batches = _generate_batches(modality_list, batch_size=transform_config.batch_size)
            else:
                modality_batches = None

            content_embeddings = _async_runner(
                filtered_content_batches,
                api_key,
                endpoint_url,
                model_name,
                transform_config.encoding_format,
                transform_config.input_type,
                transform_config.truncate,
                False,
                modalities=modality_batches,
                dimensions=dimensions,
            )
            # Build a simple row index -> embedding map
            embeddings_dict = dict(
                zip(df_content.loc[valid_content_mask].index, content_embeddings.get("embeddings", []))
            )
            info_msgs_dict = dict(
                zip(df_content.loc[valid_content_mask].index, content_embeddings.get("info_msgs", []))
            )
        else:
            embeddings_dict = {}
            info_msgs_dict = {}

        # Apply embeddings or None to all rows
        df_content = df_content.apply(_add_embeddings, embeddings=embeddings_dict, info_msgs=info_msgs_dict, axis=1)

        embedding_dataframes.append(df_content)
        content_masks.append(content_mask)

    combined_df = _concatenate_extractions_pandas(df_transform_ledger, embedding_dataframes, content_masks)

    # Embed custom content
    if custom_content_field is not None:
        result_target_field = task_config.get("result_target_field") or custom_content_field + "_embedding"

        extracted_custom_content = (
            combined_df["metadata"]
            .apply(partial(_get_pandas_custom_content, custom_content_field=custom_content_field))
            .apply(lambda x: x.strip() if isinstance(x, str) and x.strip() else None)
        )

        valid_custom_content_mask = extracted_custom_content.notna()
        if valid_custom_content_mask.any():
            custom_content_list = extracted_custom_content[valid_custom_content_mask].to_list()
            custom_content_batches = _generate_batches(custom_content_list, batch_size=transform_config.batch_size)

            custom_content_embeddings = _async_runner(
                custom_content_batches,
                api_key,
                endpoint_url,
                model_name,
                transform_config.encoding_format,
                transform_config.input_type,
                transform_config.truncate,
                False,
                dimensions=dimensions,
            )
            custom_embeddings_dict = dict(
                zip(
                    extracted_custom_content.loc[valid_custom_content_mask].index,
                    custom_content_embeddings.get("embeddings", []),
                )
            )
        else:
            custom_embeddings_dict = {}

        combined_df = combined_df.apply(
            _add_custom_embeddings, embeddings=custom_embeddings_dict, result_target_field=result_target_field, axis=1
        )

    return combined_df, {"trace_info": execution_trace_log}



# SPDX-FileCopyrightText: Copyright (c) 2024-25, NVIDIA CORPORATION & AFFILIATES.
# All rights reserved.
# SPDX-License-Identifier: Apache-2.0

import logging
from typing import Tuple, Optional, Dict, Any

import pandas as pd
from pandas import DataFrame

from . import extraction_interface_relay_constructor

from nv_ingest_api.internal.extract.pdf.pdf_extractor import extract_primitives_from_pdf_internal
from nv_ingest_api.util.exception_handlers.decorators import unified_exception_handler
from nv_ingest_api.internal.extract.docx.docx_extractor import extract_primitives_from_docx_internal
from nv_ingest_api.internal.extract.pptx.pptx_extractor import extract_primitives_from_pptx_internal
from nv_ingest_api.internal.extract.image.chart_extractor import extract_chart_data_from_image_internal
from nv_ingest_api.internal.extract.image.image_extractor import extract_primitives_from_image_internal
from nv_ingest_api.internal.extract.image.table_extractor import extract_table_data_from_image_internal
from nv_ingest_api.internal.extract.image.infographic_extractor import extract_infographic_data_from_image_internal
from nv_ingest_api.internal.schemas.extract.extract_chart_schema import ChartExtractorSchema
from nv_ingest_api.internal.schemas.extract.extract_docx_schema import DocxExtractorSchema
from nv_ingest_api.internal.schemas.extract.extract_image_schema import ImageExtractorSchema
from nv_ingest_api.internal.schemas.extract.extract_infographic_schema import (
    InfographicExtractorConfigSchema,
    InfographicExtractorSchema,
)
from nv_ingest_api.internal.schemas.extract.extract_pptx_schema import PPTXExtractorSchema
from nv_ingest_api.internal.schemas.extract.extract_table_schema import TableExtractorSchema
from nv_ingest_api.internal.schemas.meta.ingest_job_schema import (
    IngestTaskChartExtraction,
    IngestTaskTableExtraction,
)
from nv_ingest_api.internal.extract.audio.audio_extraction import extract_text_from_audio_internal
from nv_ingest_api.internal.schemas.extract.extract_audio_schema import AudioExtractorSchema

logger = logging.getLogger(__name__)


@unified_exception_handler
@extraction_interface_relay_constructor(
    api_fn=extract_primitives_from_pdf_internal,
    task_keys=["extract_text", "extract_images", "extract_tables", "extract_charts", "extract_infographics"],
)
def extract_primitives_from_pdf(
    *,
    df_extraction_ledger: pd.DataFrame,  # Ledger (e.g., a pandas DataFrame)
    extract_method: str = "pdfium",  # Determines which extraction schema to use
    extract_text: bool = True,
    extract_images: bool = True,
    extract_infographics: bool = True,
    extract_tables: bool = True,
    extract_charts: bool = True,
    text_depth: str = "page",
    # Adobe-specific parameters:
    adobe_client_id: Optional[str] = None,
    adobe_client_secret: Optional[str] = None,
    # LLama
    llama_api_key: Optional[str] = None,
    # PDFium-specific parameters:
    yolox_auth_token: Optional[str] = None,
    yolox_endpoints: Optional[Tuple[Optional[str], Optional[str]]] = None,
    yolox_infer_protocol: str = "http",
    # Nemoretriver Parse parameters:
    nemoretriever_parse_endpoints: Optional[Tuple[str, str]] = None,
    nemoretriever_parse_protocol: str = "http",
    nemoretriever_parse_model_name: str = None,
    # UnstructuredIO parameters:
    unstructured_io_api_key: Optional[str] = None,
    # Tika-specific parameter:
    tika_server_url: Optional[str] = None,
):
    """
    Extract text, images, tables, charts, and infographics from PDF documents.

    This function serves as a unified interface for PDF primitive extraction, supporting multiple
    extraction engines (pdfium, adobe, llama, nemoretriever_parse, unstructured_io, and tika).
    It processes a DataFrame containing base64-encoded PDF data and returns a new DataFrame
    with structured information about the extracted elements.

    The function uses a decorator pattern to dynamically validate configuration parameters
    and invoke the appropriate extraction pipeline. This design allows for flexible
    engine-specific configuration while maintaining a consistent interface.

    Parameters
    ----------
    df_extraction_ledger : pd.DataFrame
        DataFrame containing PDF documents to process. Must include the following columns:
        - "content" : str
            Base64-encoded PDF data
        - "source_id" : str
            Unique identifier for the document
        - "source_name" : str
            Name of the document (filename or descriptive name)
        - "document_type" : str or enum
            Document type identifier (should be "pdf" or related enum value)
        - "metadata" : Dict[str, Any]
            Dictionary containing additional metadata about the document

    extract_method : str, default "pdfium"
        The extraction engine to use. Valid options:
        - "pdfium" : PDFium-based extraction (default)
        - "adobe" : Adobe PDF Services API
        - "llama" : LlamaParse extraction
        - "nemoretriever_parse" : NVIDIA NemoRetriever Parse
        - "unstructured_io" : Unstructured.io extraction
        - "tika" : Apache Tika extraction

    extract_text : bool, default True
        Whether to extract text content from the PDFs.

    extract_images : bool, default True
        Whether to extract embedded images from the PDFs.

    extract_infographics : bool, default True
        Whether to extract infographics from the PDFs.

    extract_tables : bool, default True
        Whether to extract tables from the PDFs.

    extract_charts : bool, default True
        Whether to extract charts and graphs from the PDFs.

    text_depth : str, default "page"
        Level of text granularity to extract. Options:
        - "page" : Text extracted at page level
        - "block" : Text extracted at block level
        - "paragraph" : Text extracted at paragraph level
        - "line" : Text extracted at line level

    adobe_client_id : str, optional
        Client ID for Adobe PDF Services API. Required when extract_method="adobe".

    adobe_client_secret : str, optional
        Client secret for Adobe PDF Services API. Required when extract_method="adobe".

    llama_api_key : str, optional
        API key for LlamaParse service. Required when extract_method="llama".

    yolox_auth_token : str, optional
        Authentication token for YOLOX inference services.

    yolox_endpoints : tuple of (str, str), optional
        A tuple containing (gRPC endpoint, HTTP endpoint) for YOLOX services.
        At least one endpoint must be non-empty.

    yolox_infer_protocol : str, default "http"
        Protocol to use for YOLOX inference. Options: "http" or "grpc".

    nemoretriever_parse_endpoints : tuple of (str, str), optional
        A tuple containing (gRPC endpoint, HTTP endpoint) for NemoRetriever Parse.
        Required when extract_method="nemoretriever_parse".

    nemoretriever_parse_protocol : str, default "http"
        Protocol to use for NemoRetriever Parse. Options: "http" or "grpc".

    nemoretriever_parse_model_name : str, optional
        Model name for NemoRetriever Parse. Default is "nvidia/nemoretriever-parse".

    unstructured_io_api_key : str, optional
        API key for Unstructured.io services. Required when extract_method="unstructured_io".

    tika_server_url : str, optional
        URL for Apache Tika server. Required when extract_method="tika".

    Returns
    -------
    pandas.DataFrame
        A DataFrame containing the extracted primitives with the following columns:
        - "document_type" : Type of the extracted element (e.g., "text", "image", "table")
        - "metadata" : Dictionary containing detailed information about the extracted element
        - "uuid" : Unique identifier for the extracted element

    Raises
    ------
    ValueError
        If an unsupported extraction method is specified.
        If required parameters for the specified extraction method are missing.
        If the input DataFrame does not have the required structure.

    KeyError
        If required columns are missing from the input DataFrame.

    RuntimeError
        If extraction fails due to processing errors.

    Notes
    -----
    The function uses a decorator pattern through `extraction_interface_relay_constructor`
    which dynamically processes the parameters and validates them against the appropriate
    configuration schema. The actual extraction work is delegated to the
    `extract_primitives_from_pdf_internal` function.

    For each extraction method, specific parameters are required:
    - pdfium: yolox_endpoints
    - adobe: adobe_client_id, adobe_client_secret
    - llama: llama_api_key
    - nemoretriever_parse: nemoretriever_parse_endpoints
    - unstructured_io: unstructured_io_api_key
    - tika: tika_server_url

    Examples
    --------
    >>> import pandas as pd
    >>> import base64
    >>>
    >>> # Read a PDF file and encode it as base64
    >>> with open("document.pdf", "rb") as f:
    >>>     pdf_content = base64.b64encode(f.read()).decode("utf-8")
    >>>
    >>> # Create a DataFrame with the PDF content
    >>> df = pd.DataFrame({
    >>>     "source_id": ["doc1"],
    >>>     "source_name": ["document.pdf"],
    >>>     "content": [pdf_content],
    >>>     "document_type": ["pdf"],
    >>>     "metadata": [{"content_metadata": {"type": "document"}}]
    >>> })
    >>>
    >>> # Extract primitives using PDFium
    >>> result_df = extract_primitives_from_pdf(
    >>>     df_extraction_ledger=df,
    >>>     extract_method="pdfium",
    >>>     yolox_endpoints=(None, "http://localhost:8000/v1/infer")
    >>> )
    >>>
    >>> # Display the types of extracted elements
    >>> print(result_df["document_type"].value_counts())
    """
    pass


def extract_primitives_from_pdf_pdfium(
    df_extraction_ledger: pd.DataFrame,
    *,
    extract_text: bool = True,
    extract_images: bool = True,
    extract_tables: bool = True,
    extract_charts: bool = True,
    extract_infographics: bool = True,
    text_depth: str = "page",
    yolox_auth_token: Optional[str] = None,
    yolox_endpoints: Optional[Tuple[Optional[str], Optional[str]]] = None,
    yolox_infer_protocol: str = "http",
) -> pd.DataFrame:
    """
    Extract primitives from PDF documents using the PDFium extraction method.

    A simplified wrapper around the general extract_primitives_from_pdf function
    that defaults to using the PDFium extraction engine.

    Parameters
    ----------
    df_extraction_ledger : pd.DataFrame
        DataFrame containing PDF documents to process. Must include the following columns:
        - "content" : str
            Base64-encoded PDF data
        - "source_id" : str
            Unique identifier for the document
        - "source_name" : str
            Name of the document (filename or descriptive name)
        - "document_type" : str or enum
            Document type identifier (should be "pdf" or related enum value)
        - "metadata" : Dict[str, Any]
            Dictionary containing additional metadata about the document
    extract_text : bool, default True
        Whether to extract text content
    extract_images : bool, default True
        Whether to extract embedded images
    extract_tables : bool, default True
        Whether to extract tables
    extract_charts : bool, default True
        Whether to extract charts
    extract_infographics : bool, default True
        Whether to extract infographics
    text_depth : str, default "page"
        Level of text granularity (page, block, paragraph, line)
    yolox_auth_token : str, optional
        Authentication token for YOLOX inference services
    yolox_endpoints : tuple of (str, str), optional
        Tuple containing (gRPC endpoint, HTTP endpoint) for YOLOX services
    yolox_infer_protocol : str, default "http"
        Protocol to use for YOLOX inference ("http" or "grpc")

    Returns
    -------
    pd.DataFrame
        DataFrame containing the extracted primitives
    """
    return extract_primitives_from_pdf(
        df_extraction_ledger=df_extraction_ledger,
        extract_method="pdfium",
        extract_text=extract_text,
        extract_images=extract_images,
        extract_tables=extract_tables,
        extract_charts=extract_charts,
        extract_infographics=extract_infographics,
        text_depth=text_depth,
        yolox_auth_token=yolox_auth_token,
        yolox_endpoints=yolox_endpoints,
        yolox_infer_protocol=yolox_infer_protocol,
    )


def extract_primitives_from_pdf_nemoretriever_parse(
    df_extraction_ledger: pd.DataFrame,
    *,
    extract_text: bool = True,
    extract_images: bool = True,
    extract_tables: bool = True,
    extract_charts: bool = True,
    extract_infographics: bool = True,
    text_depth: str = "page",
    yolox_auth_token: Optional[str] = None,
    yolox_endpoints: Optional[Tuple[Optional[str], Optional[str]]] = None,
    yolox_infer_protocol: str = "http",
    nemoretriever_parse_endpoints: Optional[Tuple[str, str]] = None,
    nemoretriever_parse_protocol: str = "http",
    nemoretriever_parse_model_name: Optional[str] = None,
) -> pd.DataFrame:
    """
    Extract primitives from PDF documents using the NemoRetriever Parse extraction method.

    This function serves as a specialized wrapper around the general extract_primitives_from_pdf
    function, pre-configured to use NemoRetriever Parse as the extraction engine. It processes
    PDF documents to extract various content types including text, images, tables, charts, and
    infographics, returning the results in a structured DataFrame.

    Parameters
    ----------
    df_extraction_ledger : pd.DataFrame
        DataFrame containing PDF documents to process. Must include the following columns:
        - "content" : str
            Base64-encoded PDF data
        - "source_id" : str
            Unique identifier for the document
        - "source_name" : str
            Name of the document (filename or descriptive name)
        - "document_type" : str or enum
            Document type identifier (should be "pdf" or related enum value)
        - "metadata" : Dict[str, Any]
            Dictionary containing additional metadata about the document

    extract_text : bool, default True
        Whether to extract text content from the PDFs. When True, the function will
        attempt to extract and structure all textual content according to the
        granularity specified by `text_depth`.

    extract_images : bool, default True
        Whether to extract embedded images from the PDFs. When True, the function
        will identify, extract, and process images embedded within the document.

    extract_tables : bool, default True
        Whether to extract tables from the PDFs. When True, the function will
        detect tabular structures and convert them into structured data.

    extract_charts : bool, default True
        Whether to extract charts and graphs from the PDFs. When True, the function
        will detect and extract visual data representations.

    extract_infographics : bool, default True
        Whether to extract infographics from the PDFs. When True, the function will
        identify and extract complex visual information displays.

    text_depth : str, default "page"
        Level of text granularity to extract. Options:
        - "page" : Text extracted at page level (coarsest granularity)
        - "block" : Text extracted at block level (groups of paragraphs)
        - "paragraph" : Text extracted at paragraph level (semantic units)
        - "line" : Text extracted at line level (finest granularity)

    yolox_auth_token : Optional[str], default None
        Authentication token for YOLOX inference services used for image processing.
        Required if the YOLOX services need authentication.

    yolox_endpoints : Optional[Tuple[Optional[str], Optional[str]]], default None
        A tuple containing (gRPC endpoint, HTTP endpoint) for YOLOX services.
        Used for image processing capabilities within the extraction pipeline.
        Format: (grpc_endpoint, http_endpoint)
        Example: (None, "http://localhost:8000/v1/infer")

    yolox_infer_protocol : str, default "http"
        Protocol to use for YOLOX inference. Options:
        - "http" : Use HTTP protocol for YOLOX inference services
        - "grpc" : Use gRPC protocol for YOLOX inference services

    nemoretriever_parse_endpoints : Optional[Tuple[str, str]], default None
        A tuple containing (gRPC endpoint, HTTP endpoint) for NemoRetriever Parse.
        Format: (grpc_endpoint, http_endpoint)
        Example: (None, "http://localhost:8015/v1/chat/completions")
        Required for this extraction method.

    nemoretriever_parse_protocol : str, default "http"
        Protocol to use for NemoRetriever Parse. Options:
        - "http" : Use HTTP protocol for NemoRetriever Parse services
        - "grpc" : Use gRPC protocol for NemoRetriever Parse services

    nemoretriever_parse_model_name : Optional[str], default None
        Model name for NemoRetriever Parse.
        Default is typically "nvidia/nemoretriever-parse" if None is provided.

    Returns
    -------
    pd.DataFrame
        A DataFrame containing the extracted primitives with the following columns:
        - "document_type" : str
            Type of the extracted element (e.g., "text", "image", "structured")
        - "metadata" : Dict[str, Any]
            Dictionary containing detailed information about the extracted element
            including position, content, confidence scores, etc.
        - "uuid" : str
            Unique identifier for the extracted element

    Raises
    ------
    ValueError
        If `nemoretriever_parse_endpoints` is None or empty
        If the input DataFrame does not have the required structure

    KeyError
        If required columns are missing from the input DataFrame

    RuntimeError
        If extraction fails due to service unavailability or processing errors

    Examples
    --------
    >>> import pandas as pd
    >>> import base64
    >>>
    >>> # Read a PDF file and encode it as base64
    >>> with open("document.pdf", "rb") as f:
    >>>     pdf_content = base64.b64encode(f.read()).decode("utf-8")
    >>>
    >>> # Create a DataFrame with the PDF content
    >>> df = pd.DataFrame({
    >>>     "source_id": ["doc1"],
    >>>     "source_name": ["document.pdf"],
    >>>     "content": [pdf_content],
    >>>     "document_type": ["pdf"],
    >>>     "metadata": [{"content_metadata": {"type": "document"}}]
    >>> })
    >>>
    >>> # Extract primitives using NemoRetriever Parse
    >>> result_df = extract_primitives_from_pdf_nemoretriever_parse(
    >>>     df_extraction_ledger=df,
    >>>     nemoretriever_parse_endpoints=(None, "http://localhost:8015/v1/chat/completions")
    >>> )
    >>>
    >>> # Display the types of extracted elements
    >>> print(result_df["document_type"].value_counts())

    Notes
    -----
    - NemoRetriever Parse excels at extracting structured data like tables from PDFs
    - For optimal results, ensure both NemoRetriever Parse and YOLOX services are
      properly configured and accessible
    - The extraction quality may vary depending on the complexity and quality of the input PDF
    - This function wraps the more general `extract_primitives_from_pdf` function with
      pre-configured parameters for NemoRetriever Parse extraction
    """
    return extract_primitives_from_pdf(
        df_extraction_ledger=df_extraction_ledger,
        extract_method="nemoretriever_parse",
        extract_text=extract_text,
        extract_images=extract_images,
        extract_tables=extract_tables,
        extract_charts=extract_charts,
        extract_infographics=extract_infographics,
        text_depth=text_depth,
        yolox_endpoints=yolox_endpoints,
        yolox_auth_token=yolox_auth_token,
        yolox_infer_protocol=yolox_infer_protocol,
        nemoretriever_parse_endpoints=nemoretriever_parse_endpoints,
        nemoretriever_parse_protocol=nemoretriever_parse_protocol,
        nemoretriever_parse_model_name=nemoretriever_parse_model_name,
    )


@unified_exception_handler
def extract_primitives_from_audio(
    *,
    df_ledger: pd.DataFrame,
    audio_endpoints: Tuple[str, str],
    audio_infer_protocol: str = "grpc",
    auth_token: str = None,
    use_ssl: bool = False,
    ssl_cert: str = None,
) -> Any:
    """
    Extract audio primitives from a ledger DataFrame using the specified audio configuration.

    This function builds an extraction configuration based on the provided audio endpoints,
    inference protocol, authentication token, and SSL settings. It then delegates the extraction
    work to the internal function ``extract_text_from_audio_internal`` using the constructed
    configuration and ledger DataFrame.

    Parameters
    ----------
    df_ledger : pandas.DataFrame
        A DataFrame containing the ledger information required for audio extraction.
    audio_endpoints : Tuple[str, str]
        A tuple of two strings representing the audio service endpoints gRPC and HTTP services.
    audio_infer_protocol : str, optional
        The protocol to use for audio inference (e.g., "grpc"). Default is "grpc".
    auth_token : str, optional
        Authentication token for the audio inference service. Default is an empty string.
    use_ssl : bool, optional
        Flag indicating whether to use SSL for secure connections. Default is False.
    ssl_cert : str, optional
        Path to the SSL certificate file to use if ``use_ssl`` is True. Default is an empty string.

    Returns
    -------
    Any
        The result of the audio extraction as returned by
        ``extract_text_from_audio_internal``. The specific type depends on the internal implementation.

    Raises
    ------
    Exception
        Any exceptions raised during the extraction process will be handled by the
        ``@unified_exception_handler`` decorator.

    Examples
    --------
    >>> import pandas as pd
    >>> # Create a sample DataFrame with ledger data
    >>> df = pd.DataFrame({"audio_data": ["file1.wav", "file2.wav"]})
    >>> result = extract_primitives_from_audio(
    ...     df_ledger=df,
    ...     audio_endpoints=("http://primary.endpoint", "http://secondary.endpoint"),
    ...     audio_infer_protocol="grpc",
    ...     auth_token="secret-token",
    ...     use_ssl=True,
    ...     ssl_cert="/path/to/cert.pem"
    ... )
    """
    task_config: Dict[str, Any] = {"params": {"extract_audio_params": {}}}

    extraction_config = AudioExtractorSchema(
        **{
            "audio_extraction_config": {
                "audio_endpoints": audio_endpoints,
                "audio_infer_protocol": audio_infer_protocol,
                "auth_token": auth_token,
                "ssl_cert": ssl_cert,
                "use_ssl": use_ssl,
            }
        }
    )

    result, _ = extract_text_from_audio_internal(
        df_extraction_ledger=df_ledger,
        task_config=task_config,
        extraction_config=extraction_config,
        execution_trace_log=None,
    )

    return result


@unified_exception_handler
def extract_primitives_from_pptx(
    *,
    df_ledger: pd.DataFrame,
    extract_text: bool = True,
    extract_images: bool = True,
    extract_tables: bool = True,
    extract_charts: bool = True,
    extract_infographics: bool = True,
    yolox_endpoints: Optional[Tuple[str, str]] = None,
    yolox_infer_protocol: str = "grpc",
    auth_token: str = "",
) -> pd.DataFrame:
    """
    Extract primitives from PPTX files provided in a DataFrame.

    This function configures the PPTX extraction task by assembling a task configuration
    dictionary using the provided parameters. It then creates an extraction configuration
    object (e.g., an instance of PPTXExtractorSchema) and delegates the actual extraction
    process to the internal function `extract_primitives_from_pptx_internal`.

    Parameters
    ----------
    df_ledger : pd.DataFrame
        A DataFrame containing base64-encoded PPTX files. The DataFrame is expected to include
        columns such as "content" (with the base64-encoded PPTX) and "source_id".
    extract_text : bool, default=True
        Flag indicating whether text should be extracted from the PPTX files.
    extract_images : bool, default=True
        Flag indicating whether images should be extracted.
    extract_tables : bool, default=True
        Flag indicating whether tables should be extracted.
    extract_charts : bool, default=True
        Flag indicating whether charts should be extracted.
    extract_infographics : bool, default=True
        Flag indicating whether infographics should be extracted.
    yolox_endpoints : Optional[Tuple[str, str]], default=None
        Optional tuple containing endpoints for YOLOX inference, if needed for image analysis.
    yolox_infer_protocol : str, default="grpc"
        The protocol to use for YOLOX inference.
    auth_token : str, default=""
        Authentication token to be used with the PPTX extraction configuration.

    Returns
    -------
    pd.DataFrame
        A DataFrame containing the extracted primitives from the PPTX files. Expected columns include
        "document_type", "metadata", and "uuid".

    Notes
    -----
    This function is decorated with `@unified_exception_handler` to handle exceptions uniformly.
    The task configuration is assembled with two main keys:
      - "params": Contains boolean flags for controlling which primitives to extract.
      - "pptx_extraction_config": Contains additional settings for PPTX extraction (e.g., YOLOX endpoints,
        inference protocol, and auth token).
    It then calls `extract_primitives_from_pptx_internal` with the DataFrame, the task configuration,
    and the extraction configuration.
    """
    task_config: Dict[str, Any] = {
        "params": {
            "extract_text": extract_text,
            "extract_images": extract_images,
            "extract_tables": extract_tables,
            "extract_charts": extract_charts,
            "extract_infographics": extract_infographics,
        },
    }

    extraction_config = PPTXExtractorSchema(
        **{
            "pptx_extraction_config": {
                "yolox_endpoints": yolox_endpoints,
                "yolox_infer_protocol": yolox_infer_protocol,
                "auth_token": auth_token,
            },
        }
    )  # Assuming PPTXExtractorSchema is defined and imported

    return extract_primitives_from_pptx_internal(
        df_extraction_ledger=df_ledger,
        task_config=task_config,
        extraction_config=extraction_config,
        execution_trace_log=None,
    )


@unified_exception_handler
def extract_primitives_from_docx(
    *,
    df_ledger: pd.DataFrame,
    extract_text: bool = True,
    extract_images: bool = True,
    extract_tables: bool = True,
    extract_charts: bool = True,
    extract_infographics: bool = True,
    yolox_endpoints: Optional[Tuple[str, str]] = None,
    yolox_infer_protocol: str = "grpc",
    auth_token: str = "",
) -> pd.DataFrame:
    """
    Extract primitives from DOCX documents in a DataFrame.

    This function configures and invokes the DOCX extraction process. It builds a task configuration
    using the provided extraction flags (for text, images, tables, charts, and infographics) and additional
    settings for YOLOX endpoints, inference protocol, and authentication. It then creates a DOCX extraction
    configuration (an instance of DocxExtractorSchema) and delegates the extraction to an internal function.

    Parameters
    ----------
    df_ledger : pd.DataFrame
        The input DataFrame containing DOCX documents in base64 encoding. The DataFrame is expected to
        include required columns such as "content" (with the base64-encoded DOCX) and optionally "source_id".
    extract_text : bool, optional
        Flag indicating whether to extract text content from the DOCX documents (default is True).
    extract_images : bool, optional
        Flag indicating whether to extract images from the DOCX documents (default is True).
    extract_tables : bool, optional
        Flag indicating whether to extract tables from the DOCX documents (default is True).
    extract_charts : bool, optional
        Flag indicating whether to extract charts from the DOCX documents (default is True).
    extract_infographics : bool, optional
        Flag indicating whether to extract infographics from the DOCX documents (default is True).
    yolox_endpoints : Optional[Tuple[str, str]], optional
        A tuple containing YOLOX inference endpoints. If None, the default endpoints defined in the
        DOCX extraction configuration will be used.
    yolox_infer_protocol : str, optional
        The inference protocol to use with the YOLOX endpoints (default is "grpc").
    auth_token : str, optional
        The authentication token for accessing the YOLOX inference service (default is an empty string).

    Returns
    -------
    pd.DataFrame
        A DataFrame containing the extracted DOCX primitives. Typically, the resulting DataFrame contains
        columns such as "document_type", "metadata", and "uuid".

    Raises
    ------
    Exception
        If an error occurs during the DOCX extraction process, the exception is logged and re-raised.
    """
    # Build the task configuration with parameters and DOCX-specific extraction settings.
    task_config: Dict[str, Any] = {
        "params": {
            "extract_text": extract_text,
            "extract_images": extract_images,
            "extract_tables": extract_tables,
            "extract_charts": extract_charts,
            "extract_infographics": extract_infographics,
        },
    }

    # Create the extraction configuration object (instance of DocxExtractorSchema).
    extraction_config = DocxExtractorSchema(
        **{
            "docx_extraction_config": {
                "yolox_endpoints": yolox_endpoints,
                "yolox_infer_protocol": yolox_infer_protocol,
                "auth_token": auth_token,
            },
        }
    )

    # Delegate the actual extraction to the internal function.
    return extract_primitives_from_docx_internal(
        df_extraction_ledger=df_ledger,
        task_config=task_config,
        extraction_config=extraction_config,
        execution_trace_log=None,
    )


@unified_exception_handler
def extract_primitives_from_image(
    *,
    df_ledger: pd.DataFrame,
    extract_text: bool = True,
    extract_images: bool = True,
    extract_tables: bool = True,
    extract_charts: bool = True,
    extract_infographics: bool = True,
    yolox_endpoints: Optional[Tuple[str, str]] = None,
    yolox_infer_protocol: str = "grpc",
    auth_token: str = "",
) -> pd.DataFrame:
    task_config: Dict[str, Any] = {
        "params": {
            "extract_text": extract_text,
            "extract_images": extract_images,
            "extract_tables": extract_tables,
            "extract_charts": extract_charts,
            "extract_infographics": extract_infographics,
        },
    }

    extraction_config = ImageExtractorSchema(
        **{
            "image_extraction_config": {
                "yolox_endpoints": yolox_endpoints,
                "yolox_infer_protocol": yolox_infer_protocol,
                "auth_token": auth_token,
            },
        }
    )

    result, _ = extract_primitives_from_image_internal(
        df_extraction_ledger=df_ledger,
        task_config=task_config,
        extraction_config=extraction_config,
        execution_trace_log=None,
    )

    return result


@unified_exception_handler
def extract_chart_data_from_image(
    *,
    df_ledger: pd.DataFrame,
    yolox_endpoints: Tuple[str, str],
    ocr_endpoints: Tuple[str, str],
    yolox_protocol: str = "grpc",
    ocr_protocol: str = "grpc",
    auth_token: str = "",
) -> DataFrame:
    """
    Public interface to extract chart data from ledger DataFrame.

    Parameters
    ----------
    df_ledger : pd.DataFrame
        DataFrame containing metadata required for chart extraction.
    yolox_endpoints : Tuple[str, str]
        YOLOX inference server endpoints.
    ocr_endpoints : Tuple[str, str]
        PaddleOCR inference server endpoints.
    yolox_protocol : str, optional
        Protocol for YOLOX inference (default "grpc").
    ocr_protocol : str, optional
        Protocol for PaddleOCR inference (default "grpc").
    auth_token : str, optional
        Authentication token for inference services.
    execution_trace_log : list, optional
        Execution trace logs.

    Returns
    -------
    pd.DataFrame
        Updated DataFrame after chart extraction.

    Raises
    ------
    Exception
        If an error occurs during extraction.
    """
    task_config = IngestTaskChartExtraction()
    extraction_config = ChartExtractorSchema(
        **{
            "endpoint_config": {
                "yolox_endpoints": yolox_endpoints,
                "ocr_endpoints": ocr_endpoints,
                "yolox_infer_protocol": yolox_protocol,
                "ocr_infer_protocol": ocr_protocol,
                "auth_token": auth_token,
            }
        }
    )

    result, _ = extract_chart_data_from_image_internal(
        df_extraction_ledger=df_ledger,
        task_config=task_config,
        extraction_config=extraction_config,
        execution_trace_log=None,
    )

    return result


@unified_exception_handler
def extract_table_data_from_image(
    *,
    df_ledger: pd.DataFrame,
    yolox_endpoints: Optional[Tuple[str, str]] = None,
    ocr_endpoints: Optional[Tuple[str, str]] = None,
    yolox_protocol: Optional[str] = None,
    ocr_protocol: Optional[str] = None,
    auth_token: Optional[str] = None,
) -> pd.DataFrame:
    """
    Public interface to extract chart data from a ledger DataFrame.

    Parameters
    ----------
    df_ledger : pd.DataFrame
        DataFrame containing metadata required for chart extraction.
    yolox_endpoints : Optional[Tuple[str, str]], default=None
        YOLOX inference server endpoints. If None, the default defined in ChartExtractorConfigSchema is used.
    ocr_endpoints : Optional[Tuple[str, str]], default=None
        PaddleOCR inference server endpoints. If None, the default defined in ChartExtractorConfigSchema is used.
    yolox_protocol : Optional[str], default=None
        Protocol for YOLOX inference. If None, the default defined in ChartExtractorConfigSchema is used.
    ocr_protocol : Optional[str], default=None
        Protocol for PaddleOCR inference. If None, the default defined in ChartExtractorConfigSchema is used.
    auth_token : Optional[str], default=None
        Authentication token for inference services. If None, the default defined in ChartExtractorConfigSchema is used.

    Returns
    -------
    pd.DataFrame
        - The updated DataFrame after chart extraction.

    Raises
    ------
    Exception
        If an error occurs during extraction.
    """
    task_config = IngestTaskTableExtraction()

    config_kwargs = {
        "endpoint_config": {
            "yolox_endpoints": yolox_endpoints,
            "ocr_endpoints": ocr_endpoints,
            "yolox_infer_protocol": yolox_protocol,
            "ocr_infer_protocol": ocr_protocol,
            "auth_token": auth_token,
        }
    }
    # Remove keys with None values so that ChartExtractorConfigSchema's defaults are used.
    config_kwargs = {k: v for k, v in config_kwargs.items() if v is not None}

    extraction_config = TableExtractorSchema(**config_kwargs)

    result, _ = extract_table_data_from_image_internal(
        df_extraction_ledger=df_ledger,
        task_config=task_config,
        extraction_config=extraction_config,
        execution_trace_log=None,
    )

    return result


@unified_exception_handler
def extract_infographic_data_from_image(
    *,
    df_ledger: pd.DataFrame,
    ocr_endpoints: Optional[Tuple[str, str]] = None,
    ocr_protocol: Optional[str] = None,
    auth_token: Optional[str] = None,
) -> pd.DataFrame:
    """
    Extract infographic data from a DataFrame using the configured infographic extraction pipeline.

    This function creates a task configuration for infographic extraction, builds the extraction
    configuration from the provided PaddleOCR endpoints, protocol, and authentication token (or uses
    the default values from InfographicExtractorConfigSchema if None), and then calls the internal
    extraction function to process the DataFrame. The unified exception handler decorator ensures
    that any errors are appropriately logged and managed.

    Parameters
    ----------
    df_extraction_ledger : pd.DataFrame
        DataFrame containing the images and associated metadata from which infographic data is to be extracted.
    ocr_endpoints : Optional[Tuple[str, str]], default=None
        A tuple of PaddleOCR endpoint addresses (e.g., (gRPC_endpoint, HTTP_endpoint)) used for inference.
        If None, the default endpoints from InfographicExtractorConfigSchema are used.
    ocr_protocol : Optional[str], default=None
        The protocol (e.g., "grpc" or "http") for PaddleOCR inference.
        If None, the default protocol from InfographicExtractorConfigSchema is used.
    auth_token : Optional[str], default=None
        The authentication token required for secure access to PaddleOCR inference services.
        If None, the default value from InfographicExtractorConfigSchema is used.

    Returns
    -------
    pd.DataFrame
        The updated DataFrame after infographic extraction has been performed.

    Raises
    ------
    Exception
        Propagates any exception raised during the extraction process, after being handled by the
        unified exception handler.
    """

    task_config = {}

    extractor_config_kwargs = {
        "endpoint_config": InfographicExtractorConfigSchema(
            **{
                "ocr_endpoints": ocr_endpoints,
                "ocr_infer_protocol": ocr_protocol,
                "auth_token": auth_token,
            }
        )
    }
    # Remove keys with None values so that InfographicExtractorConfigSchema's defaults are used.
    extractor_config_kwargs = {k: v for k, v in extractor_config_kwargs.items() if v is not None}

    extraction_config = InfographicExtractorSchema(**extractor_config_kwargs)

    result, _ = extract_infographic_data_from_image_internal(
        df_extraction_ledger=df_ledger,
        task_config=task_config,
        extraction_config=extraction_config,
        execution_trace_log=None,
    )

    return result



# SPDX-FileCopyrightText: Copyright (c) 2024, NVIDIA CORPORATION & AFFILIATES.
# All rights reserved.
# SPDX-License-Identifier: Apache-2.0


# Copyright (c) 2024, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import io
import logging
import time
from typing import Any, Optional
from typing import Dict
from typing import List

import aiohttp

from nv_ingest_api.internal.enums.common import ContentTypeEnum

DEFAULT_RESULT_TYPE = "text"
DEFAULT_FILE_NAME = "_.pdf"
DEFAULT_CHECK_INTERVAL_SECONDS = 1
DEFAULT_MAX_TIMEOUT_SECONDS = 2_000

logger = logging.getLogger(__name__)


def llama_parse_extractor(
    pdf_stream: io.BytesIO,
    extract_text: bool,
    extract_images: bool,
    extract_infographics: bool,
    extract_tables: bool,
    extractor_config: dict,
    execution_trace_log: Optional[List[Any]] = None,
) -> List[Dict[ContentTypeEnum, Dict[str, Any]]]:
    """
    Helper function to use LlamaParse API to extract text from a bytestream PDF.

    Parameters
    ----------
    pdf_stream : io.BytesIO
        A bytestream PDF.
    extract_text : bool
        Specifies whether to extract text.
    extract_images : bool
        Specifies whether to extract images.
    extract_tables : bool
        Specifies whether to extract tables.
    extract_infographics : bool
        Specifies whether to extract infographics.
    extractor_config : dict
        A dictionary containing additional extraction parameters including:
            - api_key: API key for LlamaParse.
            - result_type: Type of result to extract (default provided).
            - file_name: Name of the file (default provided).
            - check_interval: Interval for checking status (default provided).
            - max_timeout: Maximum timeout in seconds (default provided).
            - row_data: Row data for additional metadata.
            - metadata_column: Column name to extract metadata (default "metadata").
    execution_trace_log : optional
        Trace information for debugging purposes.

    Returns
    -------
    List[Dict[ContentTypeEnum, Dict[str, Any]]]:
        A list of extracted data. Each item is a dictionary where the key is a
        ContentTypeEnum and the value is a dictionary containing content and metadata.

    Raises
    ------
    ValueError
        If extractor_config is not a dict or required parameters are missing.
    """

    _ = execution_trace_log  # Unused variable

    logger.debug("Extracting PDF with LlamaParse backend.")

    # Validate extractor_config.
    if not isinstance(extractor_config, dict):
        raise ValueError("extractor_config must be a dictionary.")

    api_key = extractor_config.get("llama_api_key")
    if not api_key:
        raise ValueError("LLAMA_CLOUD_API_KEY is required in extractor_config.")

    result_type = extractor_config.get("result_type", DEFAULT_RESULT_TYPE)
    file_name = extractor_config.get("file_name", DEFAULT_FILE_NAME)
    check_interval = extractor_config.get("check_interval", DEFAULT_CHECK_INTERVAL_SECONDS)
    max_timeout = extractor_config.get("max_timeout", DEFAULT_MAX_TIMEOUT_SECONDS)

    row_data = extractor_config.get("row_data")
    if row_data is None:
        raise ValueError("Missing 'row_data' in extractor_config.")
    metadata_column = extractor_config.get("metadata_column", "metadata")
    if hasattr(row_data, "index"):
        metadata = row_data[metadata_column] if metadata_column in row_data.index else {}
    else:
        metadata = row_data.get(metadata_column, {})

    extracted_data = []

    if extract_text:
        # TODO: As of Feb 2024, LlamaParse returns multi-page documents as one
        # long text. See if we can break it into pages or if LlamaParse adds
        # support for extracting each page.
        text = asyncio.run(
            async_llama_parse(
                pdf_stream,
                api_key,
                file_name=file_name,
                result_type=result_type,
                check_interval_seconds=check_interval,
                max_timeout_seconds=max_timeout,
            )
        )

        text_metadata = metadata.copy()
        text_metadata.update(
            {
                "content": text,
                "metadata": {
                    "document_type": ContentTypeEnum[result_type],
                },
            }
        )

        payload = {
            ContentTypeEnum[result_type]: text_metadata,
        }

        extracted_data.append(payload)

    # TODO: LlamaParse extracts tables, but we have to extract the tables
    # ourselves from text/markdown.
    if extract_tables:
        # Table extraction logic goes here.
        pass

    # LlamaParse does not support image extraction as of Feb 2024.
    if extract_images:
        # Image extraction logic goes here.
        pass

    # Infographics extraction is currently not supported by LlamaParse.
    if extract_infographics:
        logger.debug("Infographics extraction requested, but not supported by LlamaParse.")

    return extracted_data


async def async_llama_parse(
    pdf_stream: io.BytesIO,
    api_key: str,
    file_name: str = DEFAULT_FILE_NAME,
    result_type: str = DEFAULT_RESULT_TYPE,
    check_interval_seconds: int = DEFAULT_CHECK_INTERVAL_SECONDS,
    max_timeout_seconds: int = DEFAULT_MAX_TIMEOUT_SECONDS,
) -> str:
    """Uses the LlamaParse API to extract text from bytestream PDF.

    Parameters
    ----------
    pdf_stream : io.BytesIO
        A bytestream PDF.
    api_key: str
        API key from https://cloud.llamaindex.ai.
    file_name: str
        Name of the PDF file.
    result_type: str
        The result type for the parser. One of `text` or `markdown`.
    check_interval_seconds: int
        The interval in seconds to check if the parsing is done.
    max_timeout_seconds: int
        The maximum timeout in seconds to wait for the parsing to finish.

    Returns
    -------
    str
        A string of extracted text.
    """
    base_url = "https://api.cloud.llamaindex.ai/api/parsing"
    # Normalize in case api_key contains only whitespace; avoid sending an empty bearer token
    _token = (api_key or "").strip()
    _auth_value = f"Bearer {_token}" if _token else "Bearer <no key provided>"
    headers = {"Authorization": _auth_value}
    mime_type = "application/pdf"

    try:
        data = aiohttp.FormData()
        data.add_field(
            "file",
            pdf_stream,
            filename=file_name,
            content_type=mime_type,
        )

        upload_url = f"{base_url}/upload"

        async with aiohttp.ClientSession() as session:
            async with session.post(
                upload_url,
                data=data,
                headers=headers,
            ) as response:
                response_json = await response.json()
                job_id = response_json["id"]
                logger.debug("Started parsing the file under job_id %s" % job_id)

            result_url = f"{base_url}/job/{job_id}/result/{result_type}"

            start = time.time()
            while True:
                await asyncio.sleep(check_interval_seconds)
                result = await session.get(result_url, headers=headers)

                if result.status == 404:
                    end = time.time()
                    if end - start > max_timeout_seconds:
                        raise Exception("Timeout while parsing PDF.")
                    continue

                result_json = await result.json()
                if result.status == 400:
                    detail = result_json.get("detail", "Unknown error")
                    raise Exception(f"Failed to parse the PDF file: {detail}")

                text = result_json[result_type]
                return text

    except Exception as e:
        logger.error("Error while parsing the PDF file: ", e)
        return ""



# SPDX-FileCopyrightText: Copyright (c) 2024, NVIDIA CORPORATION & AFFILIATES.
# All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import numpy as np
from typing import List, Tuple, Optional

from nv_ingest_api.internal.primitives.nim.default_values import (
    YOLOX_MAX_BATCH_SIZE,
    YOLOX_CONF_THRESHOLD,
    YOLOX_IOU_THRESHOLD,
    YOLOX_MIN_SCORE,
    YOLOX_FINAL_SCORE,
)
from nv_ingest_api.internal.primitives.nim.model_interface.yolox import YoloxPageElementsModelInterface
from nv_ingest_api.util.image_processing.transforms import crop_image, numpy_to_base64
from nv_ingest_api.util.metadata.aggregators import CroppedImageWithContent
from nv_ingest_api.util.nim import create_inference_client

logger = logging.getLogger(__name__)


def extract_tables_and_charts_from_image(annotation_dict, original_image, page_idx, tables_and_charts):
    """
    Extract and process table and chart regions from the provided image based on detection annotations.

    Parameters
    ----------
    annotation_dict : dict
        A dictionary containing detected objects and their bounding boxes, e.g. keys "table" and "chart".
    original_image : np.ndarray
        The original image from which objects were detected.
    page_idx : int
        The index of the current page being processed.
    tables_and_charts : list of tuple
        A list to which extracted table/chart data will be appended. Each item is a tuple
        (page_idx, CroppedImageWithContent).

    Notes
    -----
    This function iterates over the detected table and chart objects. For each detected object, it:
      - Crops the original image based on the bounding box.
      - Converts the cropped image to a base64 encoded string.
      - Wraps the encoded image along with its bounding box and the image dimensions in a standardized data structure.

    Additional model inference or post-processing can be added where needed.

    Examples
    --------
    >>> annotation_dict = {"table": [ [...], [...] ], "chart": [ [...], [...] ]}
    >>> original_image = np.random.rand(1536, 1536, 3)
    >>> tables_and_charts = []
    >>> extract_tables_and_charts(annotation_dict, original_image, 0, tables_and_charts)
    """

    width, height, *_ = original_image.shape
    for label in ["table", "chart"]:
        if not annotation_dict:
            continue

        objects = annotation_dict[label]
        for idx, bboxes in enumerate(objects):
            *bbox, _ = bboxes
            h1, w1, h2, w2 = bbox

            cropped = crop_image(original_image, (int(h1), int(w1), int(h2), int(w2)))
            base64_img = numpy_to_base64(cropped)

            element_data = CroppedImageWithContent(
                content="",
                image=base64_img,
                bbox=(int(w1), int(h1), int(w2), int(h2)),
                max_width=width,
                max_height=height,
                type_string=label,
            )
            tables_and_charts.append((page_idx, element_data))


def extract_tables_and_charts_yolox(
    pages: List[Tuple[int, np.ndarray]],
    config: dict,
    trace_info: Optional[List] = None,
) -> List[Tuple[int, object]]:
    """
    Given a list of (page_index, image) tuples and a configuration dictionary,
    this function calls the YOLOX-based inference service to extract table and chart
    annotations from all pages.

    Parameters
    ----------
    pages : List[Tuple[int, np.ndarray]]
        A list of tuples containing the page index and the corresponding image.
    config : dict
        A dictionary containing configuration parameters such as:
            - 'yolox_endpoints'
            - 'auth_token'
            - 'yolox_infer_protocol'
    trace_info : Optional[List], optional
        Optional tracing information for logging/debugging purposes.

    Returns
    -------
    List[Tuple[int, object]]
        For each page, returns a tuple (page_index, joined_content) where
        joined_content is the result of combining annotations from the inference.
    """
    tables_and_charts = []
    yolox_client = None

    try:
        model_interface = YoloxPageElementsModelInterface()
        yolox_client = create_inference_client(
            config["yolox_endpoints"],
            model_interface,
            config["auth_token"],
            config["yolox_infer_protocol"],
        )

        # Collect all page indices and images in order.
        image_page_indices = [page[0] for page in pages]
        original_images = [page[1] for page in pages]

        # Prepare the data payload with all images.
        data = {"images": original_images}

        # Perform inference using the YOLOX client.
        inference_results = yolox_client.infer(
            data,
            model_name="yolox",
            max_batch_size=YOLOX_MAX_BATCH_SIZE,
            conf_thresh=YOLOX_CONF_THRESHOLD,
            iou_thresh=YOLOX_IOU_THRESHOLD,
            min_score=YOLOX_MIN_SCORE,
            final_thresh=YOLOX_FINAL_SCORE,
            trace_info=trace_info,
            stage_name="pdf_extraction",
        )

        # Process results: iterate over each image's inference output.
        for annotation_dict, page_index, original_image in zip(inference_results, image_page_indices, original_images):
            extract_tables_and_charts_from_image(
                annotation_dict,
                original_image,
                page_index,
                tables_and_charts,
            )

    except TimeoutError:
        logger.error("Timeout error during table/chart extraction.")
        raise

    except Exception as e:
        err_msg = f"Error during table/chart extraction: {str(e)}"
        logger.exception(err_msg)
        raise

    finally:
        if yolox_client:
            yolox_client.close()

    logger.debug(f"Extracted {len(tables_and_charts)} tables and charts.")
    return tables_and_charts



# SPDX-FileCopyrightText: Copyright (c) 2024, NVIDIA CORPORATION & AFFILIATES.
# All rights reserved.
# SPDX-License-Identifier: Apache-2.0

import logging
from functools import wraps
from multiprocessing import Lock
from multiprocessing import Manager

logger = logging.getLogger(__name__)

# Create a shared manager and lock for thread-safe access
manager = Manager()
global_cache = manager.dict()
lock = Lock()


def multiprocessing_cache(max_calls):
    """
    A decorator that creates a global cache shared between multiple processes.
    The cache is invalidated after `max_calls` number of accesses.

    Args:
        max_calls (int): The number of calls after which the cache is cleared.

    Returns:
        function: The decorated function with global cache and invalidation logic.
    """

    def decorator(func):
        call_count = manager.Value("i", 0)  # Shared integer for call counting

        @wraps(func)
        def wrapper(*args, **kwargs):
            key = (func.__name__, args, frozenset(kwargs.items()))

            with lock:
                call_count.value += 1

                if call_count.value > max_calls:
                    global_cache.clear()
                    call_count.value = 0

                if key in global_cache:
                    return global_cache[key]

            result = func(*args, **kwargs)

            with lock:
                global_cache[key] = result

            return result

        return wrapper

    return decorator



# SPDX-FileCopyrightText: Copyright (c) 2024, NVIDIA CORPORATION & AFFILIATES.
# All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import logging
import math
import uuid
import concurrent.futures
from typing import Any
from typing import Dict
from typing import Tuple
from typing import Optional
from typing import List

import numpy as np
import pypdfium2 as pdfium

from nv_ingest_api.internal.extract.pdf.engines.pdfium import _extract_page_elements
from nv_ingest_api.internal.primitives.nim.model_interface import nemoretriever_parse as nemoretriever_parse_utils
from nv_ingest_api.internal.enums.common import AccessLevelEnum
from nv_ingest_api.internal.enums.common import ContentTypeEnum
from nv_ingest_api.internal.enums.common import ContentDescriptionEnum
from nv_ingest_api.internal.enums.common import TableFormatEnum
from nv_ingest_api.internal.enums.common import TextTypeEnum
from nv_ingest_api.internal.schemas.meta.metadata_schema import validate_metadata
from nv_ingest_api.internal.primitives.nim.model_interface.yolox import (
    YOLOX_PAGE_IMAGE_PREPROC_WIDTH,
    YOLOX_PAGE_IMAGE_PREPROC_HEIGHT,
    YOLOX_PAGE_IMAGE_FORMAT,
)
from nv_ingest_api.internal.schemas.extract.extract_pdf_schema import NemoRetrieverParseConfigSchema
from nv_ingest_api.util.metadata.aggregators import (
    extract_pdf_metadata,
    LatexTable,
    Base64Image,
    construct_image_metadata_from_pdf_image,
    construct_text_metadata,
)
from nv_ingest_api.util.pdf.pdfium import pdfium_pages_to_numpy
from nv_ingest_api.internal.primitives.nim.default_values import YOLOX_MAX_BATCH_SIZE
from nv_ingest_api.util.exception_handlers.pdf import pdfium_exception_handler
from nv_ingest_api.util.image_processing.transforms import numpy_to_base64, crop_image
from nv_ingest_api.util.nim import create_inference_client


logger = logging.getLogger(__name__)

NEMORETRIEVER_PARSE_RENDER_DPI = 300
NEMORETRIEVER_PARSE_MAX_WIDTH = 1024
NEMORETRIEVER_PARSE_MAX_HEIGHT = 1280
NEMORETRIEVER_PARSE_MAX_BATCH_SIZE = 8


# Define a helper function to use nemoretriever_parse to extract text from a base64 encoded bytestram PDF
def nemoretriever_parse_extractor(
    pdf_stream: io.BytesIO,
    extract_text: bool,
    extract_images: bool,
    extract_infographics: bool,
    extract_tables: bool,
    extract_charts: bool,
    extractor_config: dict,
    execution_trace_log: Optional[List[Any]] = None,
) -> str:
    """
    Helper function to use nemoretriever_parse to extract text from a bytestream PDF.

    Parameters
    ----------
    pdf_stream : io.BytesIO
        A bytestream PDF.
    extract_text : bool
        Specifies whether to extract text.
    extract_images : bool
        Specifies whether to extract images.
    extract_tables : bool
        Specifies whether to extract tables.
    extract_infographics : bool
        Specifies whether to extract infographics.
    extract_charts : bool
        Specifies whether to extract charts.
    execution_trace_log : Optional[List], optional
        Trace information for debugging purposes (default is None).
    extractor_config : dict
        A dictionary containing additional extraction parameters. Expected keys include:
            - row_data : dict
            - text_depth : str, optional (default is "page")
            - extract_tables_method : str, optional (default is "yolox")
            - identify_nearby_objects : bool, optional (default is True)
            - table_output_format : str, optional (default is "pseudo_markdown")
            - pdfium_config : dict, optional (configuration for PDFium)
            - nemoretriever_parse_config : dict, optional (configuration for NemoRetrieverParse)
            - metadata_column : str, optional (default is "metadata")

    Returns
    -------
    str
        A string of extracted text.

    Raises
    ------
    ValueError
        If required keys are missing in extractor_config or invalid values are provided.
    KeyError
        If required keys are missing in row_data.
    """
    logger = logging.getLogger(__name__)
    logger.debug("Extracting PDF with nemoretriever_parse backend.")

    # Retrieve row_data from extractor_config.
    row_data = extractor_config.get("row_data")
    if row_data is None:
        raise ValueError("Missing 'row_data' in extractor_config.")

    # Get source_id from row_data.
    try:
        source_id = row_data["source_id"]
    except KeyError:
        raise KeyError("row_data must contain 'source_id'.")

    # Get and validate text_depth.
    text_depth_str = extractor_config.get("text_depth", "page")
    try:
        text_depth = TextTypeEnum[text_depth_str.upper()]
    except KeyError:
        valid_options = [e.name.lower() for e in TextTypeEnum]
        raise ValueError(f"Invalid text_depth value: {text_depth_str}. Expected one of: {valid_options}")

    # Get extraction method for tables.
    extract_tables_method = extractor_config.get("extract_tables_method", "yolox")

    # Flag for identifying nearby objects.
    identify_nearby_objects = extractor_config.get("identify_nearby_objects", True)

    # Get and validate table_output_format.
    table_output_format_str = extractor_config.get("table_output_format", "pseudo_markdown")
    try:
        table_output_format = TableFormatEnum[table_output_format_str.upper()]
    except KeyError:
        valid_options = [e.name.lower() for e in TableFormatEnum]
        raise ValueError(
            f"Invalid table_output_format value: {table_output_format_str}. Expected one of: {valid_options}"
        )

    # Process nemoretriever_parse configuration.
    nemoretriever_parse_config_raw = extractor_config.get("nemoretriever_parse_config", {})
    if isinstance(nemoretriever_parse_config_raw, dict):
        nemoretriever_parse_config = NemoRetrieverParseConfigSchema(**nemoretriever_parse_config_raw)
    elif isinstance(nemoretriever_parse_config_raw, NemoRetrieverParseConfigSchema):
        nemoretriever_parse_config = nemoretriever_parse_config_raw
    else:
        raise ValueError(
            "`nemoretriever_parse_config` must be a dictionary or a NemoRetrieverParseConfigSchema instance."
        )

    # Get base metadata.
    metadata_col = extractor_config.get("metadata_column", "metadata")
    if hasattr(row_data, "index") and metadata_col in row_data.index:
        base_unified_metadata = row_data[metadata_col]
    else:
        base_unified_metadata = row_data.get(metadata_col, {})

    # get base source_metadata
    base_source_metadata = base_unified_metadata.get("source_metadata", {})
    # get source_location
    source_location = base_source_metadata.get("source_location", "")
    # get collection_id (assuming coming in from source_metadata...)
    collection_id = base_source_metadata.get("collection_id", "")
    # get partition_id (assuming coming in from source_metadata...)
    partition_id = base_source_metadata.get("partition_id", -1)
    # get access_level (assuming coming in from source_metadata...)
    access_level = base_source_metadata.get("access_level", AccessLevelEnum.UNKNOWN)

    extracted_data = []
    doc = pdfium.PdfDocument(pdf_stream)
    pdf_metadata = extract_pdf_metadata(doc, source_id)
    page_count = pdf_metadata.page_count

    source_metadata = {
        "source_name": pdf_metadata.filename,
        "source_id": source_id,
        "source_location": source_location,
        "source_type": pdf_metadata.source_type,
        "collection_id": collection_id,
        "date_created": pdf_metadata.date_created,
        "last_modified": pdf_metadata.last_modified,
        "summary": "",
        "partition_id": partition_id,
        "access_level": access_level,
    }

    accumulated_text = []
    accumulated_tables = []
    accumulated_images = []

    pages_for_ocr = []  # We'll accumulate (page_idx, np_image) here
    pages_for_tables = []  # We'll accumulate (page_idx, np_image) here
    futures = []  # We'll keep track of all the Future objects for table/charts

    nemoretriever_parse_client = None
    if extract_text:
        nemoretriever_parse_client = _create_clients(nemoretriever_parse_config)

    max_workers = nemoretriever_parse_config.workers_per_progress_engine
    with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers) as executor:

        for page_idx in range(page_count):
            page = doc.get_page(page_idx)

            page_image, padding_offset = _convert_pdfium_page_to_numpy_for_parser(page)
            pages_for_ocr.append((page_idx, page_image))
            page_image_for_tables, padding_offset_for_tables = _convert_pdfium_page_to_numpy_for_yolox(page)
            pages_for_tables.append((page_idx, page_image_for_tables, padding_offset_for_tables))

            page.close()

            # Whenever pages_as_images hits NEMORETRIEVER_PARSE_MAX_BATCH_SIZE, submit a job
            if (extract_text) and (len(pages_for_ocr) >= NEMORETRIEVER_PARSE_MAX_BATCH_SIZE):
                future_parser = executor.submit(
                    lambda *args, **kwargs: ("parser", _extract_text_and_bounding_boxes(*args, **kwargs)),
                    pages_for_ocr[:],  # pass a copy
                    nemoretriever_parse_client,
                    execution_trace_log=execution_trace_log,
                )
                futures.append(future_parser)
                pages_for_ocr.clear()

            # Whenever pages_as_images hits YOLOX_MAX_BATCH_SIZE, submit a job
            if (
                (extract_tables_method == "yolox")
                and (extract_tables or extract_charts or extract_infographics)
                and (len(pages_for_tables) >= YOLOX_MAX_BATCH_SIZE)
            ):
                future_yolox = executor.submit(
                    lambda *args, **kwargs: ("yolox", _extract_page_elements(*args, **kwargs)),
                    pages_for_tables[:],  # pass a copy
                    page_count,
                    source_metadata,
                    base_unified_metadata,
                    extract_tables,
                    extract_charts,
                    extract_infographics,
                    table_output_format,
                    nemoretriever_parse_config.yolox_endpoints,
                    nemoretriever_parse_config.yolox_infer_protocol,
                    nemoretriever_parse_config.auth_token,
                    execution_trace_log=execution_trace_log,
                )
                futures.append(future_yolox)
                pages_for_tables.clear()

        # After page loop, if we still have leftover pages_as_images, submit one last job
        if extract_text and pages_for_ocr:
            future_parser = executor.submit(
                lambda *args, **kwargs: ("parser", _extract_text_and_bounding_boxes(*args, **kwargs)),
                pages_for_ocr[:],  # pass a copy
                nemoretriever_parse_client,
                execution_trace_log=execution_trace_log,
            )
            futures.append(future_parser)
            pages_for_ocr.clear()

        if (
            (extract_tables_method == "yolox")
            and (extract_tables or extract_charts or extract_infographics)
            and pages_for_tables
        ):
            future_yolox = executor.submit(
                lambda *args, **kwargs: ("yolox", _extract_page_elements(*args, **kwargs)),
                pages_for_tables[:],
                page_count,
                source_metadata,
                base_unified_metadata,
                extract_tables,
                extract_charts,
                extract_infographics,
                table_output_format,
                nemoretriever_parse_config.yolox_endpoints,
                nemoretriever_parse_config.yolox_infer_protocol,
                nemoretriever_parse_config.auth_token,
                execution_trace_log=execution_trace_log,
            )
            futures.append(future_yolox)
            pages_for_tables.clear()

        parser_results = []
        # Now wait for all futures to complete
        for fut in concurrent.futures.as_completed(futures):
            model_name, extracted_items = fut.result()  # blocks until finished
            if (model_name == "yolox") and (extract_tables or extract_charts or extract_infographics):
                extracted_data.extend(extracted_items)
            elif model_name == "parser":
                parser_results.extend(extracted_items)

    for page_idx, parser_output in parser_results:
        page = None
        page_image = None
        page_text = []

        page_nearby_blocks = {
            "text": {"content": [], "bbox": [], "type": []},
            "images": {"content": [], "bbox": [], "type": []},
            "structured": {"content": [], "bbox": [], "type": []},
        }

        for bbox_dict in parser_output:
            cls = bbox_dict["type"]
            bbox = bbox_dict["bbox"]
            txt = bbox_dict["text"]

            transformed_bbox = [
                math.floor(bbox["xmin"] * NEMORETRIEVER_PARSE_MAX_WIDTH),
                math.floor(bbox["ymin"] * NEMORETRIEVER_PARSE_MAX_HEIGHT),
                math.ceil(bbox["xmax"] * NEMORETRIEVER_PARSE_MAX_WIDTH),
                math.ceil(bbox["ymax"] * NEMORETRIEVER_PARSE_MAX_HEIGHT),
            ]

            if cls not in nemoretriever_parse_utils.ACCEPTED_CLASSES:
                continue

            if identify_nearby_objects:
                _insert_page_nearby_blocks(page_nearby_blocks, cls, txt, transformed_bbox)

            if extract_text:
                page_text.append(txt)

            if (extract_tables_method == "nemoretriever_parse") and (extract_tables) and (cls == "Table"):
                table = LatexTable(
                    latex=txt,
                    bbox=transformed_bbox,
                    max_width=NEMORETRIEVER_PARSE_MAX_WIDTH,
                    max_height=NEMORETRIEVER_PARSE_MAX_HEIGHT,
                )
                accumulated_tables.append(table)

            if extract_images and (cls == "Picture"):
                if page is None:
                    page = doc.get_page(page_idx)
                if page_image is None:
                    page_image, _ = _convert_pdfium_page_to_numpy_for_parser(page)

                img_numpy = crop_image(page_image, transformed_bbox)

                if img_numpy is not None:
                    base64_img = numpy_to_base64(img_numpy, format=YOLOX_PAGE_IMAGE_FORMAT)
                    image = Base64Image(
                        image=base64_img,
                        bbox=transformed_bbox,
                        width=img_numpy.shape[1],
                        height=img_numpy.shape[0],
                        max_width=NEMORETRIEVER_PARSE_MAX_WIDTH,
                        max_height=NEMORETRIEVER_PARSE_MAX_HEIGHT,
                    )
                    accumulated_images.append(image)

        # If NemoRetrieverParse fails to extract anything, fall back to using pdfium.
        if not "".join(page_text).strip():
            if page is None:
                page = doc.get_page(page_idx)
            page_text = [page.get_textpage().get_text_bounded()]

        accumulated_text.extend(page_text)

        # Construct tables
        if extract_tables:
            for table in accumulated_tables:
                extracted_data.append(
                    _construct_table_metadata(
                        table,
                        page_idx,
                        page_count,
                        source_metadata,
                        base_unified_metadata,
                    )
                )
            accumulated_tables = []

        # Construct images
        if extract_images:
            for image in accumulated_images:
                extracted_data.append(
                    construct_image_metadata_from_pdf_image(
                        image,
                        page_idx,
                        page_count,
                        source_metadata,
                        base_unified_metadata,
                    )
                )
            accumulated_images = []

        # Construct text - page
        if (extract_text) and (text_depth == TextTypeEnum.PAGE):
            extracted_data.append(
                construct_text_metadata(
                    accumulated_text,
                    pdf_metadata.keywords,
                    page_idx,
                    -1,
                    -1,
                    -1,
                    page_count,
                    text_depth,
                    source_metadata,
                    base_unified_metadata,
                    delimiter="\n\n",
                    bbox_max_dimensions=(NEMORETRIEVER_PARSE_MAX_WIDTH, NEMORETRIEVER_PARSE_MAX_HEIGHT),
                    nearby_objects=page_nearby_blocks,
                )
            )
            accumulated_text = []

    # Construct text - document
    if (extract_text) and (text_depth == TextTypeEnum.DOCUMENT):
        text_extraction = construct_text_metadata(
            accumulated_text,
            pdf_metadata.keywords,
            -1,
            -1,
            -1,
            -1,
            page_count,
            text_depth,
            source_metadata,
            base_unified_metadata,
            delimiter="\n\n",
        )

        if len(text_extraction) > 0:
            extracted_data.append(text_extraction)

    if nemoretriever_parse_client:
        nemoretriever_parse_client.close()
    doc.close()

    return extracted_data


def _extract_text_and_bounding_boxes(
    pages: list,
    nemoretriever_parse_client,
    execution_trace_log=None,
) -> list:

    # Collect all page indices and images in order.
    image_page_indices = [page[0] for page in pages]
    original_images = [page[1] for page in pages]

    # Prepare the data payload with all images.
    data = {"images": original_images}

    # Perform inference using the NimClient.
    inference_results = nemoretriever_parse_client.infer(
        data=data,
        model_name="nemoretriever_parse",
        stage_name="pdf_extraction",
        max_batch_size=NEMORETRIEVER_PARSE_MAX_BATCH_SIZE,
        execution_trace_log=execution_trace_log,
    )

    return list(zip(image_page_indices, inference_results))


def _create_clients(nemoretriever_parse_config):
    model_interface = nemoretriever_parse_utils.NemoRetrieverParseModelInterface(
        model_name=nemoretriever_parse_config.nemoretriever_parse_model_name,
    )
    nemoretriever_parse_client = create_inference_client(
        nemoretriever_parse_config.nemoretriever_parse_endpoints,
        model_interface,
        nemoretriever_parse_config.auth_token,
        nemoretriever_parse_config.nemoretriever_parse_infer_protocol,
        nemoretriever_parse_config.timeout,
    )

    return nemoretriever_parse_client


def _send_inference_request(
    nemoretriever_parse_client,
    image_array: np.ndarray,
) -> Dict[str, Any]:

    try:
        # NIM only supports processing one page at a time (batch size = 1).
        data = {"image": image_array}
        response = nemoretriever_parse_client.infer(
            data=data,
            model_name="nemoretriever_parse",
        )
    except Exception as e:
        logger.exception(f"Unhandled error during NemoRetrieverParse inference: {e}")
        raise e

    return response


def _convert_pdfium_page_to_numpy_for_parser(
    page: pdfium.PdfPage,
    render_dpi: int = NEMORETRIEVER_PARSE_RENDER_DPI,
    scale_tuple: Tuple[int, int] = (NEMORETRIEVER_PARSE_MAX_WIDTH, NEMORETRIEVER_PARSE_MAX_HEIGHT),
    padding_tuple: Tuple[int, int] = (NEMORETRIEVER_PARSE_MAX_WIDTH, NEMORETRIEVER_PARSE_MAX_HEIGHT),
) -> np.ndarray:
    page_images, padding_offsets = pdfium_pages_to_numpy(
        [page], render_dpi=render_dpi, scale_tuple=scale_tuple, padding_tuple=padding_tuple
    )

    return page_images[0], padding_offsets[0]


def _convert_pdfium_page_to_numpy_for_yolox(
    page: pdfium.PdfPage,
    scale_tuple: Tuple[int, int] = (YOLOX_PAGE_IMAGE_PREPROC_WIDTH, YOLOX_PAGE_IMAGE_PREPROC_HEIGHT),
    padding_tuple: Tuple[int, int] = (YOLOX_PAGE_IMAGE_PREPROC_WIDTH, YOLOX_PAGE_IMAGE_PREPROC_HEIGHT),
) -> np.ndarray:
    page_images, padding_offsets = pdfium_pages_to_numpy([page], scale_tuple=scale_tuple, padding_tuple=padding_tuple)

    return page_images[0], padding_offsets[0]


def _insert_page_nearby_blocks(
    page_nearby_blocks: Dict[str, Any],
    cls: str,
    txt: str,
    bbox: str,
):
    if cls in nemoretriever_parse_utils.ACCEPTED_TEXT_CLASSES:
        nearby_blocks_key = "text"
    elif cls in nemoretriever_parse_utils.ACCEPTED_TABLE_CLASSES:
        nearby_blocks_key = "structured"
    elif cls in nemoretriever_parse_utils.ACCEPTED_IMAGE_CLASSES:
        nearby_blocks_key = "images"

    page_nearby_blocks[nearby_blocks_key]["content"].append(txt)
    page_nearby_blocks[nearby_blocks_key]["bbox"].append(bbox)
    page_nearby_blocks[nearby_blocks_key]["type"].append(cls)


@pdfium_exception_handler(descriptor="nemoretriever_parse")
def _construct_table_metadata(
    table: LatexTable,
    page_idx: int,
    page_count: int,
    source_metadata: Dict,
    base_unified_metadata: Dict,
):
    content = table.latex
    table_format = TableFormatEnum.LATEX
    subtype = ContentTypeEnum.TABLE
    description = ContentDescriptionEnum.PDF_TABLE

    content_metadata = {
        "type": ContentTypeEnum.STRUCTURED,
        "description": description,
        "page_number": page_idx,
        "hierarchy": {
            "page_count": page_count,
            "page": page_idx,
            "line": -1,
            "span": -1,
        },
        "subtype": subtype,
    }
    table_metadata = {
        "caption": "",
        "table_content": content,
        "table_format": table_format,
        "table_location": table.bbox,
        "table_location_max_dimensions": (table.max_width, table.max_height),
    }
    ext_unified_metadata = base_unified_metadata.copy()

    ext_unified_metadata.update(
        {
            "content": "",
            "source_metadata": source_metadata,
            "content_metadata": content_metadata,
            "table_metadata": table_metadata,
        }
    )

    validated_unified_metadata = validate_metadata(ext_unified_metadata)

    return [ContentTypeEnum.STRUCTURED, validated_unified_metadata.model_dump(), str(uuid.uuid4())]



# SPDX-FileCopyrightText: Copyright (c) 2024, NVIDIA CORPORATION & AFFILIATES.
# All rights reserved.
# SPDX-License-Identifier: Apache-2.0


import logging

from pydantic import ValidationError

logger = logging.getLogger(__name__)


def schema_exception_handler(func, **kwargs):
    """
    A decorator that handles `ValidationError` exceptions for schema validation functions.

    This decorator wraps a function that performs schema validation using Pydantic.
    If a `ValidationError` is raised, it logs detailed error messages and raises a `ValueError` with the combined error
    messages.

    Parameters
    ----------
    func : callable
        The function to be decorated. This function is expected to perform schema validation.

    kwargs : dict
        Additional keyword arguments to be passed to the function.

    Returns
    -------
    callable
        The wrapped function that executes `func` with exception handling.

    Raises
    ------
    ValueError
        If a `ValidationError` is caught, this decorator logs the error details and raises a `ValueError` with the
        combined error messages.

    Notes
    -----
    This decorator is particularly useful for functions that validate configurations or data models,
    ensuring that any validation errors are logged and communicated clearly.

    Examples
    --------
    >>> @schema_exception_handler
    ... def validate_config(config_data):
    ...     schema = MySchema(**config_data)
    ...     return schema
    ...
    >>> try:
    ...     validate_config(invalid_config)
    ... except ValueError as e:
    ...     print(f"Caught error: {e}")
    Caught error: Invalid configuration: field1: value is not a valid integer; field2: field required
    """

    def inner_function(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except ValidationError as e:
            error_messages = "; ".join([f"{error['loc'][0]}: {error['msg']}" for error in e.errors()])
            log_error_message = f"Invalid configuration: {error_messages}"
            logger.error(log_error_message)
            raise ValueError(log_error_message)

    return inner_function



# SPDX-FileCopyrightText: Copyright (c) 2024-25, NVIDIA CORPORATION & AFFILIATES.
# All rights reserved.
# SPDX-License-Identifier: Apache-2.0

import inspect
from typing import Optional, Type, Union, Callable

from pydantic import BaseModel


def find_pydantic_config_schema(
    actor_class: Type,
    base_class_to_find: Type,
    param_name: str = "config",
) -> Optional[Type[BaseModel]]:
    """
    Introspects a class's MRO to find a Pydantic model in its __init__ signature.

    This function is designed to find the specific Pydantic configuration model
    for a pipeline actor, which might be a direct class or a proxy object.

    Parameters
    ----------
    actor_class : Type
        The actor class or proxy object to inspect.
    base_class_to_find : Type
        The specific base class (e.g., RaySource, RayStage) to look for when
        resolving the true actor class from a proxy.
    param_name : str, optional
        The name of the __init__ parameter to inspect for the Pydantic schema,
        by default "config".

    Returns
    -------
    Optional[Type[BaseModel]]
        The Pydantic BaseModel class if found, otherwise None.
    """
    # 1. Find the actual class to inspect, handling proxy objects.
    cls_to_inspect = None
    if inspect.isclass(actor_class):
        cls_to_inspect = actor_class
    else:
        for base in actor_class.__class__.__mro__:
            if inspect.isclass(base) and issubclass(base, base_class_to_find) and base is not base_class_to_find:
                cls_to_inspect = base
                break

    if not cls_to_inspect:
        return None

    # 2. Walk the MRO of the real class to find the __init__ with the typed parameter.
    for cls in cls_to_inspect.__mro__:
        if param_name in getattr(cls.__init__, "__annotations__", {}):
            try:
                init_sig = inspect.signature(cls.__init__)
                config_param = init_sig.parameters.get(param_name)
                if (
                    config_param
                    and config_param.annotation is not BaseModel
                    and issubclass(config_param.annotation, BaseModel)
                ):
                    return config_param.annotation  # Found the schema
            except (ValueError, TypeError):
                # This class's __init__ is not inspectable (e.g., a C-extension), continue up the MRO.
                continue

    return None


def find_pydantic_config_schema_for_callable(
    callable_fn: Callable,
    param_name: str = "stage_config",
) -> Optional[Type[BaseModel]]:
    """
    Introspects a callable's signature to find a Pydantic model parameter.

    This function is designed to find the specific Pydantic configuration model
    for a pipeline callable function.

    Parameters
    ----------
    callable_fn : Callable
        The callable function to inspect.
    param_name : str, optional
        The name of the parameter to inspect for the Pydantic schema,
        by default "stage_config".

    Returns
    -------
    Optional[Type[BaseModel]]
        The Pydantic BaseModel class if found, otherwise None.
    """
    try:
        sig = inspect.signature(callable_fn)
        config_param = sig.parameters.get(param_name)
        if (
            config_param
            and config_param.annotation is not BaseModel
            and hasattr(config_param.annotation, "__mro__")
            and issubclass(config_param.annotation, BaseModel)
        ):
            return config_param.annotation
    except (ValueError, TypeError):
        # Function signature is not inspectable
        pass

    return None


def find_pydantic_config_schema_unified(
    target: Union[Type, Callable],
    base_class_to_find: Optional[Type] = None,
    param_name: str = "config",
) -> Optional[Type[BaseModel]]:
    """
    Unified function to find Pydantic schema for either classes or callables.

    Parameters
    ----------
    target : Union[Type, Callable]
        The class or callable to inspect.
    base_class_to_find : Optional[Type], optional
        The specific base class to look for when resolving actor classes from proxies.
        Only used for class inspection.
    param_name : str, optional
        The name of the parameter to inspect for the Pydantic schema.
        For classes: defaults to "config"
        For callables: should be "stage_config"

    Returns
    -------
    Optional[Type[BaseModel]]
        The Pydantic BaseModel class if found, otherwise None.
    """
    if callable(target) and not inspect.isclass(target):
        # Handle callable function
        return find_pydantic_config_schema_for_callable(target, param_name)
    elif inspect.isclass(target) or hasattr(target, "__class__"):
        # Handle class or proxy object
        if base_class_to_find is None:
            # If no base class specified, we can't use the original function
            return None
        return find_pydantic_config_schema(target, base_class_to_find, param_name)
    else:
        return None
